{===========================================================================}
{                                                                           }
{      B I C - E m u l a t o r                                              }
{                                                                           }
{===========================================================================}

{ Aufruf: bicem file.com (Ausfhren)
  oder:   bicem file.com - (Ausfhren im Monitor)
  oder:   bicem file.com file.map (Monitor, MAP-File enthlt Adressen/
                                  Namen Zuordnung `$100 start')

  [Scroll] hlt Programm an (bzw. betritt Monitor), wenn der aktuelle
  Systemruf beendet ist (also evtl. [Enter] drcken). Das Programm hlt
  auch bei nicht-untersttzten Systemrufen an.

  [Alt-S] schaltet Sound an/aus, [Alt-K] schaltet Tastatur-bersetzung
  um (Wordstar -> UFO -> Xonix).

  Im Monitor: der obere Bereich zeigt Meldungen und Ablaufprotokoll an,
  in der Mitte Eingabe und Registersatz, darunter die nchsten X Befehle
  bzw. Speicherdump.
   r      laufenlassen volle Geschwindigkeit
   t      trace: Ausfhrung protokollieren
   d      dump: Anzeige Speicherdump, scrollen mit Pfeilen/PgUp/Dn
   s      Sound an/aus (wie [Alt-S])
   k      Tastatur (wie [Alt-K])
   l      Symbol-Lookup
   Enter  Eine insn ausfhren
   n      step over: eine insn, bzw. gesamte Unterroutine bei `call's
   N      step zum nchsten Befehl
   R      Breakpoint bearbeiten: hlt Programm an, wenn Speicherbereich
          gelesen / ausgefhrt wird (bzw. wenn HL reinzeigt: zum Fllen
          des M-Registers wird der Speicher gelesen)
   W      Analog fr Schreiben
   X      Analog fr Ausfhren
   A      Wie R und W zusammen
          - Breakpoint = Bereich zwischen 2 Adressen, alle Zugriffe
            dazwischen triggern den Breakpoint
          - Whrend der Eingabe [ESC] deaktiviert Breakpoint
          - BICEM kann nur je 1 Breakpoint von R, W und X.
   q      Ende

  Untersttzt werden alle 8080 Befehle und fast alle Z80 Befehle (auer
  ein paar EDxx-Opcodes zur Systemprogrammierung), einige CP/M Systemrufe
  (Bildschirmausgabe, Dateizugriff (lesen, schreiben)), und die A5105-
  Konsolen-Steuercodes. Die Geschwindigkeit auf Pentium-200 ist ungefhr
  dieselbe wie beim Original-BIC -- hier ist absolut nix optimiert.

  Dateizugriff: Es werden 2 Laufwerke a 4 MB simuliert, die das aktuelle
  Verzeichnis beinhalten (auf API-Ebene, nicht auf Sektorebene).

  Die Programme TURBO, UFO, POWER und BOWLING laufen, jedoch teilweise
  instabil.

  Statt des Programmnamens auf der Kommandozeile kann auch der Name einer
  Steuerdatei angegeben werden. Beispiel fr ein Programm, das einen
  PC 1715 voraussetzt:
    [BICEM]                            Kennung, /mu/ am Anfang stehen
    file = xonix.com                   Dateiname
    bios = $D400                       Adresse BIOS-Sprungleiste
    vram = $F800                       Video-RAM hier einblenden
    zg = comp.zg                       Zeichensatz laden
    keys = xonix                       oder Wordstar
  Standardwerte:
    bios = maximal mglicher Wert (der CP/M Einsprungpunkt sitzt immer
           an bios-3, der Sprung an [0] zeigt auf bios+3)
    vram = kein Video-RAM. Der Video-RAM des BIC ist nicht auf normalem
           Wege zugnglich. Der PC 1715 hatte Video-RAM an F800. Der
           Video-RAM ist 2000 Byte gro.
    zg =   Ldt einen entsprechenden Zeichensatz im BIC-Format (7 Bytes
           "Signatur", 2048 Zeichen Zeichensatz, Bit 0 = links (anders
           als bei der VGA)).

  (c) copyright 1999 Stefan Reuther <Streu@gmx.de>
}
{$I-}

USES CRT,Z80decode,FileIO,Cursor,DOS,StrUtil;

CONST ScreenSeg:WORD=$B800;

CONST RegA = 7;
      RegB = 0;
      RegC = 1;
      RegD = 2;
      RegE = 3;
      RegH = 4;
      RegL = 5;
      RegM = 6;

      CarryFlag    = 1;
      NegativFlag  = 2;
      PVFlag       = 4;
      HalfFlag     = 16;
      ZeroFlag     = 64;
      SignFlag     = 128;

{--- CP/M Daten ---}
CONST CPMUser:BYTE=0;
      CPMDrive:BYTE=0;
      CPMDTA:WORD=$80;
      CPMwriteprot:WORD=0;

VAR CPMdpb:WORD;        { DPB (einer fr alle) }
    CPMdphbase:WORD;    { Basisadresse DPH-Array }
    CPMscratch:WORD;    { 128-bytes Scratchpad }
    CPMalv:WORD;        { Allocation Vector (einer fr alle) }

VAR sys:WORD;           { Adresse fr Systemcode }

CONST KeyTrans:(kt_Wordstar, kt_Ufo, kt_Xonix, kt_LAST) = kt_Wordstar;
      InitScreen:BOOLEAN = FALSE;    { nach dem Programmende Bildschirm neu init? }

{--- Hauptspeicher ---}
TYPE TMemory = ARRAY[0..65520] OF BYTE;

     TBreakpoint = RECORD
            At0, At1 : WORD;
            Enabled  : BOOLEAN;
     END;

{--- Virtuelle Maschine ---}
VAR Regs:ARRAY[0..7] OF BYTE;     { Registersatz }
    iff:BOOLEAN;
    Shadow:ARRAY[0..5] OF BYTE;   { Schattenregister }
    shadowAF:WORD;
    Flags:BYTE;                   { Flags }
    IX,IY:WORD;                   { Indexregister }
    PC:WORD;
    SP:WORD;
    M:^TMemory;                   { Hauptspeicher }
    brk:WORD;                     { Obergrenze }

{--- Zeitgeber ---}
VAR timer_enabled  : BOOLEAN;     { Timer aktiv? }
    timer_interval : INTEGER;     { Intervall (18 = 1 Sekunde) }
    timer_call     : WORD;        { Zieladresse }
    timer_tick     : LONGINT;     { Wann kommt der nchste Timer? }

    pc_timer       : LONGINT ABSOLUTE $0040:$006C;

{--- Eingabe ---}
    con_pending : BOOLEAN;        { liegt ein Zeichen an? }
    con_char    : BYTE;           { wenn ja, welches }

{--- Monitor ---}
CONST bRead  : TBreakpoint = (At0:0; At1:0; Enabled: FALSE);
      bWrite : TBreakpoint = (At0:0; At1:0; Enabled: FALSE);
      bIO    : TBreakpoint = (At0:0; At1:0; Enabled: FALSE);
      bExec  : TBreakpoint = (At0:0; At1:0; Enabled: FALSE);

      VRAM   : TBreakpoint = (At0:0; At1:0; Enabled: FALSE);

      monitoring : BOOLEAN = FALSE;

VAR broken : BOOLEAN;             { TRUE = angehalten }

PROCEDURE MonLog(s:STRING; y:BYTE); FORWARD;

{---------------------------------------------------------------------------}
{ Hilfsfunktionen                                                           }
{---------------------------------------------------------------------------}
CONST Hex:ARRAY[0..15] OF CHAR='0123456789ABCDEF';

TYPE Str7=STRING[7];

FUNCTION HexByte(y:BYTE):Str7;
BEGIN
  HexByte := Hex[y DIV 16] + Hex[y MOD 16];
END;

FUNCTION HexWord(y:WORD):Str7;
BEGIN
  HexWord := HexByte(y DIV 256) + HexByte(y MOD 256);
END;

PROCEDURE DoHalt;
BEGIN
  CursorOn;
  IF Initscreen THEN Textmode(co80);
  Halt;
END;

{===========================================================================}
{                                                                           }
{   Konsolen-Emulation                                                      }
{                                                                           }
{===========================================================================}

CONST ConStatus : (c_normal,c_esc,c_gotoxy,c_escesc,c_esc5E,c_esc5F) = c_normal;
      ConAttr   : BYTE    = 7;
      ConY      : BYTE    = 0;          { fr Status c_gotoxy }
      ConSound  : BOOLEAN = TRUE;       { Beeper erlaubt? }

{---------------------------------------------------------------------------}
{ Tastenbersetzung ndern (Alt-K)                                          }
{---------------------------------------------------------------------------}
PROCEDURE ChangeKeyTrans;
VAR S:STRING[15];
BEGIN
  Inc(KeyTrans);
  IF KeyTrans=kt_LAST THEN KeyTrans:=kt_Wordstar;
  CASE KeyTrans OF
    kt_Wordstar: S:='WordStar';
    kt_Ufo: S:='UFO';
    kt_Xonix: S:='Xonix';
  ELSE S:=HexByte(Ord(KeyTrans));
  END;
  MonLog('Key translation set to ' + S, 9);
END;

{---------------------------------------------------------------------------}
{ Sound an/aus (Alt-S)                                                      }
{---------------------------------------------------------------------------}
PROCEDURE ChangeSound;
BEGIN
  ConSound:=NOT ConSound;
  IF ConSound THEN MonLog('Speaker turned on', 9) ELSE MonLog('Speaker turned off', 9);
END;

{---------------------------------------------------------------------------}
{ Konsolenausgabe                                                           }
{---------------------------------------------------------------------------}
PROCEDURE Con(c:BYTE);
VAR i,j,k:INTEGER;
    L:LONGINT;
CONST Frames:ARRAY[0..15] OF CHAR='ڿ´ĳ12345';
BEGIN
  TextAttr := ConAttr;
  CASE ConStatus OF
   {:: ESC ESC '0'..'9' text NUL
    :: Funktionstaste belegen }
    c_escesc:IF c=0 THEN ConStatus:=c_normal;
   {:: ESC '^' 010v0abi
    :: Invers, wenn v-Bit gesetzt, b = blinkend, i = intensiv
    :: a dient der Zeichensatzumschaltung (US<->DE) }
    c_esc5E:BEGIN
      IF c AND $10=0 THEN ConAttr:=7 ELSE ConAttr:=$70;
      IF c AND 1<>0 THEN ConAttr:=ConAttr OR 15;
      IF c AND 2<>0 THEN ConAttr:=ConAttr OR $80;
      ConStatus:=c_normal;
    END;
   {:: ESC '_' 01zzzzbi
    :: Grafikzeichen ausgeben, blinkend oder intensiv }
    c_esc5F:BEGIN
      IF c AND $40<>0 THEN BEGIN
        IF c AND 1<>0 THEN TextAttr:=TextAttr OR 15;
        IF c AND 2<>0 THEN TextAttr:=TextAttr OR $80;
        Write(Frames[(c SHR 2) AND 15]);
      END;
      ConStatus:=c_normal;
    END;
   {:: ESC 1yyyyyyy 1xxxxxxx
    :: Gehe zu X/Y, 0-based }
    c_gotoxy:BEGIN
      GotoXY((c AND $7F)+1, ConY+1);
      ConStatus:=c_normal;
    END;
   {:: ESC - Allgemeiner Steuercode }
    c_esc:CASE c OF
      $80..$FF:BEGIN
        ConStatus:=c_gotoxy;
        ConY:=c AND $7F;
      END;
      27:ConStatus:=c_escesc;
      $5E:ConStatus:=c_esc5E;
      $5F:ConStatus:=c_esc5F;
    END;
   {:: weitere Steuerzeichen }
    c_normal:CASE c OF
      0:;                                               { padding }
      1:GotoXY(1,1);                                    { home }
      7:BEGIN                                           { bel }
        IF ConSound THEN Sound(880);
        L:=MemL[$40:$6C];
        REPEAT UNTIL MemL[$40:$6C]-L>2;
        NoSound;
      END;
      9:REPEAT Write(' ') UNTIL WhereX AND 7=1;         { ^I tab }
      8:IF WhereX>1 THEN GotoXY(WhereX-1,WhereY);       { ^H bs }
      12:ClrScr;                                        { ^L ff }
      16:;                                              { ^P Druckerecho an/aus }
      19:REPEAT UNTIL Readkey<>#0;                      { ^S Pause }
      20:BEGIN                                          { ^T clreos }
        ClrEOL;
        j:=WhereY;
        k:=WhereX;
        FOR i:=j+1 TO 24 DO BEGIN
          GotoXY(1,i);
          ClrEOL;
        END;
        GotoXY(k,j);
      END;
      21:IF WhereX<80 THEN GotoXY(WhereX+1,WhereY);     { ^U Cursor rechts }
      22:ClrEOL;                                        { ^V clreol }
      24:BEGIN                                          { ^X Zeile lschen }
        GotoXY(1,WhereY);
        ClrEOL;
      END;
      26:IF WhereY>1 THEN GotoXY(WhereX, WhereY-1);     { ^Z Zeile hoch }
      27:ConStatus:=c_esc;                              { ESC - Steuercode }
      $82:CursorOn;
      $83:CursorOff;
      $84:ConAttr := 7;
      $85:ConAttr := $70;
      $86:ConAttr := 15;
      $87:ConAttr := $7F;
    ELSE Write(Chr(c));
    END;
  END;
END;

PROCEDURE ProcessInput;
VAR x,y:BYTE;
BEGIN
  y:=Ord(Readkey);
  IF y=0 THEN BEGIN
    x:=Ord(Readkey);
    CASE x OF
      83:y:=127;              { Del }
      31:ChangeSound;         { Alt-S }
      37:ChangeKeyTrans;      { Alt-K }
    ELSE CASE KeyTrans OF
      kt_Ufo:CASE x OF
        75:y:=Ord(',');       { Links }
        77:y:=Ord('/');       { Rechts }
        80:y:=Ord('.');       { Unten }
      END;
      kt_Xonix:CASE x OF
        75:y:=Ord('4');       { Links }
        77:y:=Ord('6');       { Rechts }
        80:y:=Ord('5');       { Unten }
        72:y:=Ord('8');       { Oben }
      END;
      kt_Wordstar:CASE x OF
        75:y:=Ord(^S);        { Links }
        77:y:=Ord(^D);        { Rechts }
        80:y:=Ord(^X);        { Unten }
        72:y:=Ord(^E);        { Oben }
        73:y:=Ord(^R);        { Home }
        81:y:=Ord(^C);        { PgUp }
        115:y:=Ord(^A);       { Ctrl-Links }
        116:y:=Ord(^F);       { Ctrl-Rechts }
      END;
    END; END;
  END;
  IF y<>0 THEN BEGIN
    con_pending := TRUE;
    con_char := y;
  END ELSE
   con_pending := FALSE;
END;

{---------------------------------------------------------------------------}
{ Konsoleneingabe                                                           }
{---------------------------------------------------------------------------}
FUNCTION ConIn:BYTE;
BEGIN
  WHILE NOT con_pending DO ProcessInput;
  ConIn:=con_char;
  con_pending:=FALSE;
END;

{ Liefert TRUE, wenn Konsoleneingabe anliegt }
FUNCTION ConStat:BOOLEAN;
BEGIN
  IF NOT con_pending THEN
   WHILE Keypressed AND NOT con_pending DO ProcessInput;
  ConStat := con_pending;
END;

{===========================================================================}
{                                                                           }
{   Peripheriezugriff                                                       }
{                                                                           }
{===========================================================================}
{---------------------------------------------------------------------------}
{ Hauptspeicher lesen                                                       }
{---------------------------------------------------------------------------}
FUNCTION ReadByte(adr:WORD):BYTE; FAR;
BEGIN
  IF bRead.Enabled AND (adr>=bRead.At0) AND (adr<=bRead.At1) AND NOT broken THEN BEGIN
    MonLog('`Read'' breakpoint hit.',$4E);
    broken:=TRUE;
  END;
  IF adr>=brk THEN
   ReadByte := 0
  ELSE IF VRAM.Enabled AND (adr>=VRAM.At0) AND (adr<=VRAM.At1) THEN
    ReadByte := Mem[ScreenSeg:(adr - VRAM.At0) * 2]
  ELSE
   ReadByte := M^[adr];
END;

FUNCTION ReadWord(adr:WORD):WORD;
BEGIN
  ReadWord := ReadByte(adr) + 256*ReadByte(adr+1);
END;

{---------------------------------------------------------------------------}
{ Hauptspeicher schreiben                                                   }
{---------------------------------------------------------------------------}
PROCEDURE PutByte(adr:WORD; y:BYTE);
BEGIN
  IF bWrite.Enabled AND (adr>=bWrite.At0) AND (adr<=bWrite.At1) AND NOT broken THEN BEGIN
    MonLog('`Write'' breakpoint hit.',$4E);
    broken:=TRUE;
  END;
  IF adr<brk THEN M^[adr] := y;
  IF VRAM.Enabled AND (adr>=VRAM.At0) AND (adr<=VRAM.At1) THEN BEGIN
    Mem[$B800:2*(Adr-VRAM.At0)] := y;
  END;
END;

PROCEDURE PutWord(adr:WORD; y:WORD);
BEGIN
  PutByte(adr, Lo(y));
  PutByte(adr+1, Hi(y));
END;

{---------------------------------------------------------------------------}
{ Befehle lesen                                                             }
{---------------------------------------------------------------------------}
FUNCTION Fetch:BYTE;
BEGIN
  Fetch := ReadByte(PC);
  Inc(PC);
END;

FUNCTION FetchS:SHORTINT;
BEGIN
  FetchS:=Fetch;
END;

FUNCTION FetchWord:WORD;
BEGIN
  FetchWord := ReadWord(PC);
  Inc(PC,2);
END;

{---------------------------------------------------------------------------}
{ Ports. Zustzlich zu der Adresse /adr/ nutzen einige Hacks noch aus, da  }
{ der Z80 diverse Register auf den Bus legt:                                }
{  IN/OUT n    A8..A15 = A, A0..A7 = n                                      }
{  IN/OUT r    A8..A15 = B, A0..A7 = C                                      }
{---------------------------------------------------------------------------}
FUNCTION InByte(adr:BYTE):BYTE;
BEGIN
  IF adr=$70 THEN BEGIN
    IF ConSound THEN InByte:=1 ELSE InByte:=0;
  END ELSE BEGIN
    MonLog('Unsupported port ' + HexByte(adr) + ' read',$70);
    InByte:=0;
  END;
END;

PROCEDURE OutByte(adr:BYTE; y:BYTE);
BEGIN
  IF adr=$70 THEN ConSound:=Odd(y) ELSE BEGIN
    MonLog('Unsupported port ' + HexByte(adr) + ' <- ' + HexByte(y),$70);
  END;
END;

{===========================================================================}
{                                                                           }
{   M o n i t o r                                                           }
{                                                                           }
{===========================================================================}

CONST MonLogLines = 12;                { Anzahl Protokollzeilen }
      MonBottom = MonLogLines + 4;     { Beginn Dump/Insn-Fenster }
      MonBottomLines = 26 - MonBottom; { Zeilen im Dump/Insn-Fenster }

VAR MonitorLog : ARRAY[1..MonLogLines] OF STRING[80];
    MonitorCol : ARRAY[1..MonLogLines] OF BYTE;

PROCEDURE InitMonitor;
BEGIN
  FillChar(MonitorLog, Sizeof(MonitorLog), 0);
  FillChar(MonitorCol, Sizeof(MonitorCol), 0);
END;

{---------------------------------------------------------------------------}
{ In Zeile Z (1..25) des Monitorbereiches schreiben, y = Attribut           }
{---------------------------------------------------------------------------}
PROCEDURE MonWrite(z:INTEGER; S:STRING; y:BYTE);
VAR i,j:INTEGER;
    w:WORD;
BEGIN
  Adr := (24+z) * 160;
  j:=Length(S);
  IF j>80 THEN j:=80;
  FOR i:=1 TO j DO BEGIN
    Mem[ScreenSeg:Adr] := Ord(S[i]);
    Inc(Adr);
    Mem[ScreenSeg:Adr] := y;
    Inc(Adr);
  END;
  FOR i:=j+1 TO 80 DO BEGIN
    w:=256*y;
    IF S[1]='' THEN Inc(w, 176);
    MemW[ScreenSeg:Adr] := w;
    Inc(Adr,2);
  END;
END;

{---------------------------------------------------------------------------}
{ Zeile in den scrollenden Log-Bereich des Monitors schreiben               }
{---------------------------------------------------------------------------}
PROCEDURE MonLog(s:STRING; y:BYTE);
VAR i:INTEGER;
BEGIN
  IF Monitoring THEN BEGIN
    FOR i:=1 TO MonLogLines-1 DO BEGIN
      MonitorLog[i] := MonitorLog[i+1];
      MonitorCol[i] := MonitorCol[i+1];
      MonWrite(i, MonitorLog[i], MonitorCol[i]);
    END;
    MonitorLog[MonLogLines] := S;
    MonitorCol[MonLogLines] := y;
    MonWrite(MonLogLines, S, y);
  END ELSE BEGIN
    MonWrite(0, ' ' + s + ' ', y);
  END;
END;

{---------------------------------------------------------------------------}
{ Breakpoint /b/ bearbeiten, berschrift /Name/                             }
{---------------------------------------------------------------------------}
PROCEDURE EditBreak(VAR b:TBreakpoint; Name:STRING);
VAR Adr1,Adr2:WORD;
    ok:BOOLEAN;
    c:CHAR;
BEGIN
  Adr1:=b.At0;
  ok:=FALSE;
  REPEAT
    MonWrite(MonLogLines+1, 'Enter '+Name+' breakpoint start, [ESC] for none: ' + Hexword(Adr1) + '_', $30);
    c:=Upcase(Readkey);
    CASE c OF
      #27:BEGIN
        b.Enabled:=FALSE;
        Exit;
      END;
      #13:ok:=TRUE;
      '0'..'9':Adr1:=16*Adr1 + Ord(c) - 48;
      'A'..'F':Adr1:=16*Adr1 + Ord(c) - 55;
    END;
  UNTIL ok;
  Adr2:=Adr1;
  REPEAT
    MonWrite(MonLogLines+1, 'Enter '+Name+' breakpoint end, [ESC] for none: ' + Hexword(Adr2) + '_', $30);
    c:=Upcase(Readkey);
    CASE c OF
      #27:BEGIN
        b.Enabled:=FALSE;
        Exit;
      END;
      #13:ok:=TRUE;
      '0'..'9':Adr2:=16*Adr2 + Ord(c) - 48;
      'A'..'F':Adr2:=16*Adr2 + Ord(c) - 55;
    END;
  UNTIL ok;
  b.At0:=Adr1;
  b.At1:=Adr2;
  b.Enabled:=TRUE;
END;

{---------------------------------------------------------------------------}
{ Adresse fr Speicherdump eingeben, ret /true/ wenn OK                     }
{---------------------------------------------------------------------------}
FUNCTION Dump(VAR aAdr:WORD):BOOLEAN;
VAR ok:BOOLEAN;
    c:CHAR;
    Adr:WORD;
BEGIN
  ok:=FALSE;
  Adr:=SP;
  Dump:=FALSE;
  REPEAT
    MonWrite(MonLogLines+1, 'Enter address for dump: ' + Hexword(Adr) + '_', $30);
    c:=Upcase(Readkey);
    CASE c OF
      #27:Exit;
      #13:ok:=TRUE;
      '0'..'9':Adr:=16*Adr + Ord(c) - 48;
      'A'..'F':Adr:=16*Adr + Ord(c) - 55;
      'P':Adr:=PC;
    END;
  UNTIL ok;
  Dump:=TRUE;
  aAdr:=Adr;
END;

{---------------------------------------------------------------------------}
{ Speicherdump fr Adresse /Adr/ ausgeben                                   }
{---------------------------------------------------------------------------}
PROCEDURE PrintDump(Adr:WORD);
VAR S:STRING;
    i,j:INTEGER;
BEGIN
  FOR i:=MonBottom TO 24 DO BEGIN
    S:=HexWord(Adr) + ':  ';
    FOR j:=0 TO 15 DO BEGIN
      S:=S+HexByte(ReadByte(Adr+j))+' ';
      IF j=7 THEN S:=S+' ';
    END;
    S:=S+' ';
    FOR j:=0 TO 15 DO S:=S+Chr(ReadByte(Adr+j));
    Inc(Adr,16);
    MonWrite(i,S,9);
  END;
END;

PROCEDURE LookupSymbol;
VAR S:STRING;
    p:PMap;
    c:CHAR;
BEGIN
  S:='';
  REPEAT
    p:=Map;
    WHILE (p<>NIL) AND (UpStr(S)<>UpStr(p^.Name)) DO p:=p^.Next;
    IF p<>NIL THEN
     MonWrite(MonLogLines+1, 'Symbol: ' + p^.Name + '_ is at ' + HexWord(p^.Ofs + SymbolOffset), $30)
    ELSE
     MonWrite(MonLogLines+1, 'Symbol: ' + S + '_', $30);
    c:=Upcase(Readkey);
    CASE c OF
      #27,#13:Exit;
      #25:S:='';
      #8:IF S<>'' THEN Dec(S[0]);
      #33..#127:S:=S+c;
    END;
  UNTIL FALSE;
END;

{---------------------------------------------------------------------------}
{ Monitor-Routine. Wird nach jedem Maschinenbefehl aufgerufen               }
{---------------------------------------------------------------------------}
PROCEDURE Monitor;
CONST RegNames:ARRAY[0..7] OF CHAR='BCDEHLMA';
      FlagNames:ARRAY[0..7,BOOLEAN] OF CHAR=(('c', 'C'),
                                         ('+', '-'),
                                         ('p', 'P'),
                                         ('.', '*'),
                                         ('h', 'H'),
                                         ('.', '*'),
                                         ('z', 'Z'),
                                         ('s', 'S'));
CONST quiet : BOOLEAN = FALSE;
      stepping : BOOLEAN = FALSE;
      stepTo : WORD=0;
      tracing : BOOLEAN = FALSE;
      dumpadr : WORD = 0;
VAR w,w0:WORD;
    c:CHAR;
    S:STRING;
    S1:STRING[10];
    i:INTEGER;
    Next:WORD;
    can_step:BOOLEAN;
    cont:BOOLEAN;
    dumping:BOOLEAN;
    timer_dif:LONGINT;
BEGIN
 {--- Monitor inaktiv - erlaube nur Abbruch mit [Scroll] ---}
  IF NOT Monitoring THEN BEGIN
    IF broken THEN BEGIN
      Mem[$40:$17] := Mem[$40:$17] AND NOT 16;
      GotoXY(1,25);
      Writeln;
      DoHalt;
    END;
    Exit;
  END;
 {--- verschiedene Haltebedingungen ---}
  IF bExec.Enabled AND (PC>=bExec.At0) AND (PC<=bExec.At1) THEN BEGIN
    MonLog('`Exec'' breakpoint hit.',$4E);
    broken:=TRUE;
  END;
  IF broken OR (Mem[$40:$17] AND 16<>0) OR (Stepping AND (PC=stepTo)) THEN quiet:=FALSE;
  IF NOT quiet THEN BEGIN
   {--- Anhalten: "Monitor betreten" ---}
    IF broken THEN tracing:=FALSE;
    broken:=FALSE;
    dumping:=FALSE;
    Mem[$40:$17] := Mem[$40:$17] AND NOT 16;

   { Logging abgelaufener Insns }
    w := DecodeInsn(PC, ReadByte);
    Next:=W;
    can_step := NOT ((Disasm[1]='j') OR (Disasm[1]='r'));
    IF Disasm<>'nop' THEN MonLog(HexWord(PC) + ': ' + Code + ' ' + Disasm, 7);

   { Register und Folgebefehle }
    S:='';
    FOR i:=0 TO 7 DO S := S + ' ' + RegNames[i] + '='+HexByte(Regs[i]);
    S:=S+' SP='+HexWord(SP)+' PC='+HexWord(PC)+' Flags=';
    FOR i:=0 TO 7 DO S:=S+FlagNames[i, Flags AND (1 SHL i)<>0];
    IF iff THEN S:=S+' ei' ELSE S:=S+' di';
    MonWrite(MonLogLines + 2, S, $2F);
    S:=' IX='+HexWord(IX)+' IY='+HexWord(IY)+' (SP)='+HexWord(ReadWord(SP));
    IF timer_enabled THEN BEGIN
      Str(timer_tick - pc_timer, S1);
      S:=S + '  timer:' + S1;
    END;
    MonWrite(MonLogLines + 3, S, $2F);
    FOR i:=1 TO MonBottomLines DO BEGIN
      w0:=w;
      w:=DecodeInsn(w,ReadByte);
      MonWrite(MonBottom + i - 1, HexWord(w0) + ': ' + Code + ' ' + Disasm,7);
    END;
    cont:=FALSE;
   { Interaktion }
    IF NOT tracing THEN REPEAT
      MonWrite(MonLogLines + 1, 'Monitor: [r] run [q] quit [n] step over [t] trace [RWXA] break [d] dump', $2E);
      timer_dif := timer_tick - pc_timer;
      c:=Readkey;
      CASE c OF
        #0:CASE Readkey OF
          #72:IF dumping THEN BEGIN Dec(DumpAdr,16); PrintDump(dumpadr); END;
          #73:IF dumping THEN BEGIN Dec(DumpAdr,16*MonBottomLines); PrintDump(dumpadr); END;
          #80:IF dumping THEN BEGIN Inc(DumpAdr,16); PrintDump(dumpadr); END;
          #81:IF dumping THEN BEGIN Inc(DumpAdr,16*MonBottomLines); PrintDump(dumpadr); END;
        END;
        'd':IF Dump(dumpadr) THEN BEGIN
          Dumping:=TRUE;
          PrintDump(dumpadr);
        END;
        't':BEGIN
          cont:=TRUE;
          quiet:=FALSE;
          tracing:=TRUE;
        END;
        's':ChangeSound;
        'r':BEGIN
          quiet:=TRUE;
          stepping:=FALSE;
          cont:=TRUE;
        END;
        'l':LookupSymbol;
        'k':ChangeKeyTrans;
        'q':DoHalt;
        'n','N':BEGIN
          IF can_step OR (c='N') THEN BEGIN
            stepping:=TRUE;
            stepto:=Next;
            quiet:=TRUE;
          END;
          cont:=TRUE;
        END;
        'R':EditBreak(bRead, 'Read');
        'W':EditBreak(bWrite,'Write');
        'X':EditBreak(bExec, 'Exec');
        'A':BEGIN
          EditBreak(bRead, 'Access');
          bWrite:=bRead;
        END;
        ' ',#13:cont:=TRUE;
      END;
      timer_tick := timer_dif + pc_timer;
    UNTIL cont;
    MonWrite(MonLogLines + 1, 'running...', 8);
  END;
END;

{===========================================================================}
{                                                                           }
{   Laden des auszufhrenden Programms                                      }
{                                                                           }
{===========================================================================}
PROCEDURE Strip(VAR S:STRING);
BEGIN
  WHILE (S<>'') AND (S[1]<=' ') DO Delete(S,1,1);
  WHILE (S<>'') AND (S[Length(S)]<=' ') DO Dec(S[0]);
END;

{---------------------------------------------------------------------------}
{ Laden des Zeichensatzes                                                   }
{---------------------------------------------------------------------------}
PROCEDURE LoadZG(S:STRING);
VAR F:FILE;
    buf:ARRAY[0..256*16-1] OF BYTE;
    rbuf:ARRAY[0..7] OF BYTE;
    i,j,k:INTEGER;
BEGIN
  Assign(F,S);
  Reset(F,1);
  IF IOResult<>0 THEN BEGIN
    Writeln('Could not load character set.');
    Exit;
  END;
  Seek(F,7);
  FOR i:=0 TO 255 DO BEGIN
    Blockread(F,rbuf,8);
    FOR j:=0 TO 7 DO BEGIN
      rbuf[j] := ((rbuf[j] SHR 4) AND $0F) OR ((rbuf[j] SHL 4) AND $F0);
      rbuf[j] := ((rbuf[j] SHR 2) AND $33) OR ((rbuf[j] SHL 2) AND $CC);
      rbuf[j] := ((rbuf[j] SHR 1) AND $55) OR ((rbuf[j] SHL 1) AND $AA);
      IF Monitoring THEN buf[8*i+j] := rbuf[j] ELSE BEGIN
        buf[16*i+2*j] := rbuf[j];
        buf[16*i+2*j+1] := rbuf[j];
      END;
    END;
  END;
  Close(F);
  InitScreen:=TRUE;

  ASM
        push    bp
        mov     ax,1100h
        mov     cx,256
        mov     dx,0
        mov     bh,16
        mov     bl,0
        cmp     Monitoring,0
        je      @@ok
        mov     bh,8
  @@ok: push    ss
        pop     es
        lea     bp,buf
        int     10h
        pop     bp
  END;
END;

{---------------------------------------------------------------------------}
{ Ldt das auszufhrende Programm                                           }
{---------------------------------------------------------------------------}
PROCEDURE LoadImage;
VAR R,w:WORD;
    F:FILE;
    FT:TEXT;
    Name:STRING;
    S,S1:STRING;
    i,j:INTEGER;
TYPE Sign=ARRAY[1..7] OF CHAR;
BEGIN
  FileMode:=0;
  Name := ParamStr(1);
  REPEAT
    Assign(F, Name);
    Reset(F,1);
    IF IOResult<>0 THEN BEGIN
      Writeln('Could not load image ', ParamStr(1),': exiting.');
      Halt;
    END;
    Blockread(F, m^[$100], brk-$100, R);
    Close(F);
    Name:='';
   {=== Spezialbehandlung fr Konfigurationsdateien ===}
    IF Sign((@m^[$100])^)='[BICEM]' THEN BEGIN
      S:='';
      w:=0;
      m^[$100+R] := 10;
      WHILE w<=R DO BEGIN
        IF m^[$100+w] = 10 THEN BEGIN
          i:=Pos('=',S);
          IF i<>0 THEN BEGIN
            S1:=Copy(S,1,i-1);
            Delete(S,1,i);
            Strip(S);
            Strip(S1);
            FOR i:=1 TO Length(S1) DO S1[i]:=Upcase(S1[i]);
            IF (S1<>'') AND (S1[1]=';') THEN ELSE
            IF (Length(S1)=2) AND (S1[1]>='A') AND (S1[1]<='P') AND (S1[2]=':') THEN BEGIN
              MountDrive(Ord(S1[1])-Ord('A'),S);
            END ELSE
            IF S1='FILE' THEN Name:=S ELSE
            IF S1='BIOS' THEN BEGIN
              Val(S, i, j);
              IF j=0 THEN sys:=i-3 ELSE Writeln('Invalid BIOS specification `',S,'''.');
            END ELSE IF S1='VRAM' THEN BEGIN
              Val(S, i, j);
              IF j=0 THEN BEGIN
                VRAM.At0 := i;
                VRAM.At1 := i + 80*25-1;
                VRAM.Enabled:=TRUE;
              END ELSE Writeln('Invalid VRAM specification `',S,'''.');
            END ELSE IF S1='ZG' THEN LoadZG(S) ELSE
            IF S1='KEYS' THEN BEGIN
              FOR i:=1 TO Length(S) DO S[i]:=Upcase(S[i]);
              IF S='UFO' THEN KeyTrans:=kt_UFO ELSE
              IF S='XONIX' THEN Keytrans:=kt_Xonix ELSE
               Keytrans:=kt_Wordstar;
            END ELSE
            IF S1='MESSAGE' THEN BEGIN
              TextAttr:=15;
              Writeln(S);
              TextAttr:=7;
            END ELSE
            Writeln('Parameter `',S1,''' ignored.');
          END;
          S:='';
        END ELSE IF m^[$100+w]<>13 THEN S:=S+Chr(m^[$100+w]);
        Inc(w);
      END;
      IF Name='' THEN BEGIN
        Writeln('No name specified in config file (`file=program.com'')');
        Halt;
      END;
    END;
  UNTIL Name='';
 {=== MAP laden ===}
  IF (ParamCount>1) AND (ParamStr(2)<>'-') THEN BEGIN
    Assign(FT,ParamStr(2));
    Reset(FT);
    IF IOResult<>0 THEN BEGIN
      Writeln('Could not load MAP file: symbols will be unavailable.');
    END ELSE BEGIN
      LoadMap(FT);
      Close(FT);
    END;
  END;
END;

{===========================================================================}
{                                                                           }
{   CPU Emulation                                                           }
{                                                                           }
{===========================================================================}

{---------------------------------------------------------------------------}
{ Registerzugriff                                                           }
{---------------------------------------------------------------------------}
PROCEDURE SetBC(w:WORD);
BEGIN
  Regs[RegB]:=Hi(w);
  Regs[RegC]:=Lo(w);
END;

PROCEDURE SetDE(w:WORD);
BEGIN
  Regs[RegD]:=Hi(w);
  Regs[RegE]:=Lo(w);
END;

PROCEDURE SetHL(w:WORD);
BEGIN
  Regs[RegH]:=Hi(w);
  Regs[RegL]:=Lo(w);
END;

PROCEDURE SetAF(w:WORD);
BEGIN
  Regs[RegA]:=Hi(w);
  Flags:=Lo(w);
END;

FUNCTION AF:WORD; ASSEMBLER;
ASM
        mov     ah,Regs[RegA].byte
        mov     al,Flags
END;

FUNCTION BC:WORD; ASSEMBLER;
ASM
        mov     al,Regs[RegC].byte
        mov     ah,Regs[RegB].byte
END;

FUNCTION DE:WORD; ASSEMBLER;
ASM
        mov     al,Regs[RegE].byte
        mov     ah,Regs[RegD].byte
END;

FUNCTION HL:WORD; ASSEMBLER;
ASM
        mov     al,Regs[RegL].byte
        mov     ah,Regs[RegH].byte
END;

PROCEDURE SetFlag(flag:BYTE; onoff:BOOLEAN);
BEGIN
  IF onoff THEN Flags := Flags OR flag ELSE
   Flags := Flags AND NOT flag;
END;

{---------------------------------------------------------------------------}
{ Doppelregister-Arithmetik                                                 }
{---------------------------------------------------------------------------}
PROCEDURE AddDbl(VAR hl:WORD; value:WORD);
VAR althl,neuhl:WORD;
BEGIN
  Flags := Flags AND NOT NegativFlag;
  althl := HL;
  neuhl := althl + value;
  hl := neuhl;
  SetFlag(CarryFlag, althl > neuhl);
  SetFlag(ZeroFlag, neuhl = 0);
  SetFlag(PVFlag, ((value<$8000) AND (althl<$8000) AND (neuhl>=$8000)) OR
                  ((value>=$8000) AND (althl>=$8000) AND (neuhl<$8000)));
END;

PROCEDURE AddHL(value:WORD);
VAR v:WORD;
BEGIN
  v := HL;
  AddDbl(v, value);
  SetHL(v);
END;

PROCEDURE AdcHL(value:WORD);
BEGIN
  IF Flags AND CarryFlag<>0 THEN Inc(value);
  AddHL(value);
END;

PROCEDURE SbbHL(value:WORD);
VAR althl,neuhl:WORD;
BEGIN
  IF Flags AND CarryFlag<>0 THEN Inc(value);
  Flags := Flags OR NegativFlag;
  althl := HL;
  neuhl := althl - value;
  SetHL(neuhl);
  SetFlag(CarryFlag, value > althl);
  SetFlag(ZeroFlag, neuhl = 0);
  SetFlag(PVFlag, ((value<$8000) AND (althl>=$8000) AND (neuhl<$8000)) OR
                  ((value>=$8000) AND (althl<$8000) AND (neuhl>=$8000)));
END;

{---------------------------------------------------------------------------}
{ Stack-Primitive                                                           }
{---------------------------------------------------------------------------}
PROCEDURE Push(w:WORD);
BEGIN
  Dec(SP,2);
  PutWord(SP, w);
END;

PROCEDURE Call(w:WORD);
BEGIN
  Push(PC);
  PC:=w;
END;

FUNCTION Pop:WORD;
BEGIN
  Pop:=ReadWord(SP);
  Inc(SP,2);
END;

{---------------------------------------------------------------------------}
{ ALU Emulation                                                             }
{---------------------------------------------------------------------------}
{ Sign/Zero setzen }
PROCEDURE SetSZ(reg:BYTE);
BEGIN
  SetFlag(Zeroflag, reg=0);
  SetFlag(Signflag, reg>=$80);
END;

{ Flags nach Logikoperation setzen }
PROCEDURE SetLog(L:BYTE);
BEGIN
  Flags := Flags AND NOT (PVflag OR Carryflag OR Negativflag);
  L:=L XOR (L SHR 4);
  L:=L XOR (L SHR 2);
  L:=L XOR (L SHR 1);
  IF NOT Odd(L) THEN Inc(Flags, PVflag);
END;

{---------------------------------------------------------------------------}
{ Arithmetik auf dem Akku, Operand 2 = op2, Code = Insn-Code                }
{---------------------------------------------------------------------------}
PROCEDURE Arit(code, op2:BYTE);
VAR op1:WORD;
BEGIN
  op1 := Regs[RegA];
  CASE code OF
   {--- add, adc ---}
    0,1:BEGIN
      op1 := op1 + op2;
      IF (code=1) AND (Flags AND Carryflag<>0) THEN Inc(op1);
      Flags := Flags AND NOT NegativFlag;
      SetFlag(Carryflag, op1>$FF);
      SetFlag(Halfflag, (op1 AND $F0) <> (Regs[RegA] + op2) AND $F0);
      SetFlag(PVFlag, op1 AND $80<>0);
      SetSZ(op1);
    END;
   {--- sub, sbc, cmp ---}
    2,3,7:BEGIN
      op1 := op1 - op2;
      IF (code=3) AND (Flags AND Carryflag<>0) THEN Dec(op1);
      Flags := Flags OR NegativFlag;
      SetFlag(Carryflag, op1>$FF);
      SetFlag(Halfflag, (op1 AND $F0) <> (Regs[RegA] - op2) AND $F0);
      SetFlag(PVFlag, op1 AND $80<>0);
      SetSZ(op1);
      IF code=7 THEN op1 := Regs[RegA];
    END;
   {--- and ---}
    4:BEGIN
      op1 := op1 AND op2;
      Flags := Flags OR HalfFlag;
      SetLog(op1);
      SetSZ(op1);
    END;
   {--- xor ---}
    5:BEGIN
      op1 := op1 XOR op2;
      Flags := Flags AND NOT HalfFlag;
      SetLog(op1);
      SetSZ(op1);
    END;
   {--- or ---}
    6:BEGIN
      op1 := op1 OR op2;
      Flags := Flags AND NOT HalfFlag;
      SetLog(op1);
      SetSZ(op1);
    END;
  END;
  Regs[RegA] := op1;
END;

{ Flags nach Shiftoperation setzen }
PROCEDURE SetShift(L:BYTE);
BEGIN
  Flags := Flags AND NOT (PVflag OR Negativflag OR HalfFlag);
  L:=L XOR (L SHR 4);
  L:=L XOR (L SHR 2);
  L:=L XOR (L SHR 1);
  IF NOT Odd(L) THEN Inc(Flags, PVflag);
END;

{---------------------------------------------------------------------------}
{ Shifts                                                                    }
{---------------------------------------------------------------------------}
PROCEDURE DoShiftOp(code:BYTE; VAR op2:BYTE);
VAR op1:BYTE;
    f:BYTE;
BEGIN
  op1 := op2;
  CASE code OF
   {--- rlc ---}
    0:BEGIN
      SetFlag(CarryFlag, op1 AND $80<>0);
      op1 := 2*op1;
      IF Flags AND CarryFlag<>0 THEN Inc(op1);
    END;
   {--- rrc ---}
    1:BEGIN
      SetFlag(CarryFlag, Odd(op1));
      op1 := op1 SHR 1;
      IF Flags AND CarryFlag<>0 THEN Inc(op1, $80);
    END;
   {--- rl ---}
    2:BEGIN
      IF Flags AND CarryFlag<>0 THEN f:=1 ELSE f:=0;
      SetFlag(CarryFlag, op1 AND $80<>0);
      op1 := 2*op1 + f;
    END;
   {--- rr ---}
    3:BEGIN
      IF Flags AND CarryFlag<>0 THEN f:=$80 ELSE f:=0;
      SetFlag(CarryFlag, Odd(op1));
      op1 := op1 SHR 1 + f;
    END;
   {--- sla ---}
    4:BEGIN
      SetFlag(CarryFlag, op1 AND $80<>0);
      op1 := 2*op1;
    END;
   {--- sra ---}
    5:BEGIN
      SetFlag(CarryFlag, Odd(op1));
      op1 := (op1 SHR 1) OR (op1 AND $80);
    END;
   {--- srl ---}
    7:BEGIN
      SetFlag(CarryFlag, Odd(op1));
      op1 := (op1 SHR 1);
    END;
  ELSE
    MonLog('Unsupported shift code ' + HexByte(code), $70);
    broken:=TRUE;
    Exit;
  END;
  op2 := op1;
  Flags := Flags AND NOT (NegativFlag OR HalfFlag);
  SetShift(op1);
  SetSZ(op1);
END;

{---------------------------------------------------------------------------}
{ 8-bit DEC                                                                 }
{---------------------------------------------------------------------------}
PROCEDURE Dec8(VAR reg:BYTE); ASSEMBLER;
ASM
        mov     ah,Flags
        sahf
        les     di,reg
        dec     byte ptr es:[di]
        lahf
        or      ah,NegativFlag
        mov     Flags,ah
END;

{---------------------------------------------------------------------------}
{ 8-bit INC                                                                 }
{---------------------------------------------------------------------------}
PROCEDURE Inc8(VAR reg:BYTE); ASSEMBLER;
ASM
        mov     ah,Flags
        sahf
        les     di,reg
        inc     byte ptr es:[di]
        lahf
        and     ah,NOT NegativFlag
        mov     Flags,ah
END;

{---------------------------------------------------------------------------}
{ DAA                                                                       }
{---------------------------------------------------------------------------}
PROCEDURE DecimalAdjust; ASSEMBLER;
ASM
        mov     ah,Flags
        sahf
        mov     al,Regs[RegA].byte
        test    ah,NegativFlag
        jz      @@Plus
        das
        jmp     @@ok
@@Plus: daa
@@ok:   lahf
        mov     Flags,ah
        mov     Regs[RegA].byte, al
END;

{===========================================================================}
{                                                                           }
{   CP/M Emulation                                                          }
{                                                                           }
{===========================================================================}

VAR readdir_r:SearchRec;           { aktives FindFirst }
CONST readdir_drive:INTEGER=-1;    { dessen Laufwerksbuchstabe }

{---------------------------------------------------------------------------}
{ Dateinamen /Name/ an Adresse /Adr/ poken:                                 }
{  1 byte readdir_drive                                                     }
{ 11 byte Dateiname                                                         }
{ + Padding bis 32 Byte                                                     }
{---------------------------------------------------------------------------}
PROCEDURE PokeFilename(Name:STRING; Adr:WORD);
VAR i,j:INTEGER;
BEGIN
  PutByte(Adr, readdir_drive); Inc(Adr);
  i:=1;
  FOR j:=1 TO 8 DO BEGIN
    IF (i<=Length(Name)) AND (Name[i]<>'.') THEN BEGIN
      PutByte(Adr, Ord(Name[i]));
      Inc(i);
    END ELSE PutByte(Adr, 32);
    Inc(Adr);
  END;
  WHILE (i<=Length(Name)) AND (Name[i]<>'.') DO Inc(i);
  Inc(i);
  FOR j:=1 TO 3 DO BEGIN
    IF (i<=Length(Name)) THEN BEGIN
      PutByte(Adr, Ord(Name[i]));
      Inc(i);
    END ELSE PutByte(Adr, 32);
    Inc(Adr);
  END;
  FOR i:=12 TO 32 DO BEGIN
    PutByte(Adr, 0);
    Inc(Adr);
  END;
END;

{---------------------------------------------------------------------------}
{ func=0 -> create                                                          }
{ func=1 -> open                                                            }
{ func=2 -> erase                                                           }
{ func=3 -> findfirst                                                       }
{ func=4 -> compute file size                                               }
{---------------------------------------------------------------------------}
PROCEDURE UseFCB(func:BYTE);
VAR Adr:WORD;
    Drive:BYTE;
    Name:STRING[15];
    pf:PFile;
    i:INTEGER;
    F:FILE;
    L:LONGINT;
BEGIN
  Adr:=DE;
  Drive := ReadByte(Adr);
  IF Drive=0 THEN Drive:=CPMDrive ELSE Dec(Drive);
  Name:='';
  FOR i:=1 TO 8 DO Name:=Name + Chr(ReadByte(Adr+i) AND $7F);
  WHILE Name[Length(Name)]=' ' DO Dec(Name[0]);
  Name:=Name + '.';
  FOR i:=9 TO 11 DO Name:=Name + Chr(ReadByte(Adr+i) AND $7F);
  WHILE Name[Length(Name)]=' ' DO Dec(Name[0]);

  CASE func OF
    0:BEGIN
      Regs[RegA] := OpenFile(Drive, Name, PF, TRUE);
      IF Regs[RegA]=0 THEN BEGIN
        PutByte(Adr+12, 0);              { Extent-Nummer }
        PutWord(Adr+13, pf^.Index);      { `internal use' }
        PutByte(Adr+15, 0);              { Record-Nummer }
      END;
    END;
    1:BEGIN
      Regs[RegA] := OpenFile(Drive, Name, PF, FALSE);
      IF Regs[RegA]=0 THEN BEGIN
        PutByte(Adr+12, 0);              { Extent-Nummer }
        PutWord(Adr+13, pf^.Index);      { `internal use' }
        PutByte(Adr+15, 0);              { Record-Nummer }
      END;
    END;
    2:IF Odd(DrivesMounted SHR Drive) THEN BEGIN
      Assign(F, DrivePrefix[Drive] + Name);
      Erase(F);
      IF IOResult<>0 THEN BEGIN
        Regs[RegA] := 255;
      END ELSE BEGIN
        Regs[RegA] := 0;
      END;
    END ELSE Regs[RegA]:=255;
    3:IF Odd(DrivesMounted SHR Drive) THEN BEGIN
      FindFirst(DrivePrefix[Drive] + Name, AnyFile, readdir_r);
      WHILE (DOSError=0) AND (readdir_r.Attr AND (Directory OR VolumeID) <> 0) DO
       FindNext(readdir_r);
      IF DOSError=0 THEN BEGIN
        PokeFilename(readdir_r.Name, CPMDTA);
        Regs[RegA] := 0;
        readdir_drive := Drive;
      END ELSE BEGIN
        Regs[RegA] := $FF;
        readdir_drive := -1;
      END;
    END ELSE BEGIN
      Regs[RegA]:=$FF;
      readdir_drive := -1;
    END;
    4:IF Odd(DrivesMounted SHR Drive) THEN BEGIN
      Regs[RegA] := OpenFile(Drive, Name, PF, TRUE);
      IF Regs[RegA]=0 THEN BEGIN
        L:=(pf^.Size + $7F) DIV $80;
        PutWord(Adr + 33, L);
        PutByte(Adr + 35, L SHR 16);
        CloseFile(pf^.Index);
      END;
    END;
  END;
END;

{---------------------------------------------------------------------------}
{ Verschiedene FCB-Funktionen                                               }
{---------------------------------------------------------------------------}
PROCEDURE ReadSeqFCB;
VAR Adr:WORD;
    extent,rec:BYTE;
    index:WORD;
    block:ARRAY[0..127] OF BYTE;
    i:INTEGER;
BEGIN
  Adr:=DE;
  extent:=ReadByte(Adr + 12);
  index:=ReadWord(Adr + 13);
  rec:=ReadByte(Adr + 15);
  Regs[RegA] := ReadRecord(index, extent, rec, block);
  IF Regs[RegA]=0 THEN BEGIN
    FOR i:=0 TO 127 DO PutByte(CPMDTA + i, block[i]);
    Inc(rec);
    IF rec>=$80 THEN BEGIN
      Rec:=0;
      Inc(extent);
      PutByte(Adr + 12, extent);
    END;
    PutByte(Adr + 15, rec);
  END;
END;

PROCEDURE WriteSeqFCB;
VAR Adr:WORD;
    extent,rec:BYTE;
    index:WORD;
    block:ARRAY[0..127] OF BYTE;
    i:INTEGER;
BEGIN
  Adr:=DE;
  extent:=ReadByte(Adr + 12);
  index:=ReadWord(Adr + 13);
  rec:=ReadByte(Adr + 15);
  FOR i:=0 TO 127 DO block[i] := ReadByte(CPMDTA + i);
  Regs[RegA] := WriteRecord(index, extent, rec, block);
  IF Regs[RegA]=0 THEN BEGIN
    Inc(rec);
    IF rec>=$80 THEN BEGIN
      Rec:=0;
      Inc(extent);
      PutByte(Adr + 12, extent);
    END;
    PutByte(Adr + 15, rec);
  END;
END;

PROCEDURE RandomRead;
VAR Adr:WORD;
    extent,rec:BYTE;
    index:WORD;
    block:ARRAY[0..127] OF BYTE;
    i:INTEGER;
    ofs:LONGINT;
BEGIN
  Adr:=DE;
  index:=ReadWord(Adr + 13);
  ofs := ReadWord(Adr + 33);
  rec := ofs AND 127;           { 128 Stze / Extent }
  extent := ofs DIV 128;
  Regs[RegA] := ReadRecord(index, extent, rec, block);
  IF Regs[RegA]=0 THEN BEGIN
    FOR i:=0 TO 127 DO PutByte(CPMDTA + i, block[i]);
  END;
END;

PROCEDURE RandomWrite;
VAR Adr:WORD;
    extent,rec:BYTE;
    index:WORD;
    block:ARRAY[0..127] OF BYTE;
    i:INTEGER;
    ofs:LONGINT;
BEGIN
  Adr:=DE;
  index:=ReadWord(Adr + 13);
  ofs := ReadWord(Adr + 33);
  rec := ofs AND 127;           { 128 Stze / Extent }
  extent := ofs DIV 128;
  FOR i:=0 TO 127 DO block[i] := ReadByte(CPMDTA + i);
  Regs[RegA] := WriteRecord(index, extent, rec, block);
END;

{---------------------------------------------------------------------------}
{ CP/M Systemruf (call 5)                                                   }
{---------------------------------------------------------------------------}
PROCEDURE Syscall;
VAR w:WORD;
    S:STRING;
    y:BYTE;
BEGIN
  IF Monitoring THEN MonWrite(MonLogLines + 1, 'syscall active',8);
  CASE Regs[RegC] OF
   {--- 0 - boot ---}
    0:BEGIN
      MonLog('Program ends (BDOS Function 0).', $1E);
      broken:=TRUE;
      Dec(PC,2);
      Exit;
    END;
   {--- 1 = read con mit Echo ---}
    1:BEGIN
      Regs[RegA] := ConIn;
      Con(Regs[RegA]);
    END;
   {--- 2 = write con ---}
    2:Con(Regs[RegE]);
   {--- 3 = read aux ---}
    3:Regs[RegA] := 0;
   {--- 4,5 = write aux,prn ---}
    4,5:;
   {--- 6 = direct con I/O ---}
    6:IF Regs[RegE]=$FF THEN BEGIN
      IF ConStat THEN BEGIN
        Regs[RegA] := ConIn;
      END ELSE BEGIN
        Regs[RegA]:=0;
      END;
    END ELSE Con(Regs[RegE]);
   {--- 7,8 - I/O byte lesen/schreiben ---}
    7:Regs[RegA] := 0;
    8:;
   {--- 9 - Textausgabe ---}
    9:BEGIN
      w:=DE;
      WHILE ReadByte(w)<>Ord('$') DO BEGIN
        Con(ReadByte(w));
        Inc(w);
      END;
    END;
   {--- 10 - Texteingabe ---}
    10:BEGIN
      Readln(S);
      y:=ReadByte(DE)-1;
      IF Length(S) > y THEN S[0] := Chr(y);
      {S:=S+#13;}
      FOR w:=0 TO Length(S) DO PutByte(DE+1+w, Ord(S[w]));
    END;
   {--- 11 - Test, ob Eingabe anliegt ---}
    11:IF ConStat THEN Regs[RegA]:=1 ELSE Regs[RegA]:=0;
   {--- 12 - Versionscode ---}
    12:SetHL($0020);          { CP/M 2.0 }
   {--- 13 - reset disk ---}
    13:BEGIN
      Regs[RegA]:=0;
      CPMDrive:=0;
      CPMwriteprot:=0;
    END;
   {--- 14 - select disk ---}
    14:IF Odd(DrivesMounted SHR Regs[RegE]) THEN CPMDrive:=Regs[RegE];
   {--- 15 - open FCB ---}
    15:UseFCB(1);
   {--- 16 - close fcb ---}
    16:Regs[RegA] := CloseFile(ReadWord(DE + 13));
   {--- 17,18 - Find File ---}
    17:UseFCB(3);
    18:BEGIN
      IF readdir_drive<>-1 THEN BEGIN
        FindNext(readdir_r);
        WHILE (DOSError=0) AND (readdir_r.Attr AND (Directory OR VolumeID) <> 0) DO
         FindNext(readdir_r);
        IF DOSError<>0 THEN readdir_drive:=-1;
      END;
      IF readdir_drive<>-1 THEN BEGIN
        Regs[RegA] := 0;
        PokeFilename(readdir_r.Name, CPMDTA);
      END ELSE BEGIN
        Regs[RegA] := $FF;
      END;
    END;
   {--- 19 - Datei lschen ---}
    19:UseFCB(2);
   {--- 20 - read sequential ---}
    20:ReadSeqFCB;
   {--- 21 - write sequential ---}
    21:WriteSeqFCB;
   {--- 22 - create FCB ---}
    22:UseFCB(0);

   {--- 24 - Laufwerksvektor ---}
    24:SetHL(DrivesMounted);
   {--- 25 - Abfrage gesetztes LW ---}
    25:Regs[RegA] := CPMDrive;
   {--- 26 - DTA setzen ---}
    26:CPMDTA:=DE;
   {--- 27 - ALV holen ---}
    27:SetHL(ReadWord(CPMdphbase + 14 + 16*CPMdrive));
   {--- 28 - Dieses LW schtzen ---}
    28:CPMwriteprot:=CPMwriteprot OR (1 SHL CPMdrive);
   {--- 29 - Schreibgeschtzte LW ---}
    29:SetHL(CPMwriteprot);
   {--- 31 - Get DPB ---}
    31:SetHL(CPMDPB);
   {--- 32 - User-Bereich setzen ---}
    32:IF Regs[RegE]=$FF THEN Regs[RegA] := CPMUser ELSE CPMUser:=Regs[RegE];
   {--- 33 - Random Read ---}
    33:RandomRead;
   {--- 34 - Random Write ---}
    34:RandomWrite;
   {--- 35 - Compute File Size ---}
    35:UseFCB(4);
  ELSE
    MonLog('Unknown CP/M call ' + HexByte(Regs[RegC]), $70);
    broken:=TRUE;
  END;
  PC:=Pop;
  MonWrite(MonLogLines + 1, 'running...', 8);
END;

{---------------------------------------------------------------------------}
{ Laufwerk /drive/ mit dem Hostpfad verbinden, der als nullterminierter     }
{ String an DE steht                                                        }
{---------------------------------------------------------------------------}
PROCEDURE MountDriveDE(Drive:BYTE);
VAR W:WORD;
    S:STRING;
    c:BYTE;
BEGIN
  S:='';
  W:=DE;
  REPEAT
    c:=ReadByte(W);
    Inc(w);
    IF c<>0 THEN S:=S+Chr(c);
  UNTIL c=0;
  MountDrive(Drive, S);
END;

{---------------------------------------------------------------------------}
{ CP/M BIOS Ruf                                                             }
{---------------------------------------------------------------------------}
PROCEDURE BiosCall;
VAR c:BYTE;
    i:INTEGER;
    w:WORD;
BEGIN
  IF Monitoring THEN MonWrite(MonLogLines + 1, 'BIOS active',8);
  c:=Fetch;
  CASE c OF
    2:IF ConStat THEN Regs[RegA] := $FF ELSE Regs[RegA] := 0;
    3:Regs[RegA] := ConIn;
    4:Con(Regs[RegC]);
    5,6:{list out, punch out};
    7:Regs[RegA] := 0;
    8:{seek 0 auf LW};
    9:IF Regs[RegC]<CPMNumDrives THEN SetHL(CPMdphbase + 16*Regs[RegC])
     ELSE SetHL(0);
    10:{Spur whlen};
    11:{Sektor whlen};
    16:SetHL(BC); {Sektor bersetzen}
    128:BEGIN
      timer_enabled  := TRUE;
      timer_call     := DE;
      timer_interval := BC;
      timer_tick     := 0;
    END;
    129:timer_enabled := FALSE;
    130:BEGIN
      SetFlag(CarryFlag, timer_enabled);
      IF timer_enabled THEN BEGIN
        SetBC(timer_interval);
        SetDE(timer_call);
      END;
    END;
    131:BEGIN
      SetFlag(CarryFlag, VRAM.Enabled);
      IF VRAM.Enabled THEN SetBC(VRAM.At0);
      SetDE(sys);
      SetHL(brk);
    END;
    132:IF (Regs[RegC]<CPMNumDrives) THEN MountDriveDE(Regs[RegC]);
    133:IF Regs[RegC]<CPMNumDrives THEN BEGIN
      UnmountDrive(Regs[RegC]);
      IF CPMDrive=Regs[RegC] THEN CPMDrive:=0;
    END;
    134:IF (Regs[RegC]<CPMNumDrives) AND Odd(DrivesMounted SHR Regs[RegC]) THEN BEGIN
      SetFlag(CarryFlag,FALSE);
      w:=DE;
      FOR i:=1 TO Length(DrivePrefix[Regs[RegC]]) DO BEGIN
        PutByte(w, Ord(DrivePrefix[Regs[RegC]][i]));
        Inc(w);
      END;
      PutByte(w, 0);
    END ELSE SetFlag(CarryFlag,TRUE);
  ELSE
    MonLog('Unknown BIOS call ' + HexByte(c), $70);
    broken:=TRUE;
  END;
  PC:=Pop;
  MonWrite(MonLogLines + 1, 'running...', 8);
END;

{---------------------------------------------------------------------------}
{ Opcodes 00??????b                                                         }
{---------------------------------------------------------------------------}
PROCEDURE Handle0(opc:BYTE);
VAR y:SHORTINT;
    w:WORD;
BEGIN
  CASE opc OF
    $00:;
    $01:SetBC(FetchWord);
    $02:PutByte(BC, Regs[RegA]);
    $03:SetBC(BC+1);
    $04:Inc8(Regs[0]);
    $05:Dec8(Regs[0]);
    $06:Regs[0] := Fetch;
    $07:DoShiftOp(0, Regs[RegA]);
    $08:BEGIN w:=AF; SetAF(ShadowAF); ShadowAF:=w; END;
    $09:AddHL(BC);
    $0A:Regs[RegA] := ReadWord(BC);
    $0B:SetBC(BC-1);
    $0C:Inc8(Regs[1]);
    $0D:Dec8(Regs[1]);
    $0E:Regs[1] := Fetch;
    $0F:DoShiftOp(1, Regs[RegA]);
    $10:BEGIN Dec8(Regs[RegB]); y:=Fetch; IF Flags AND ZeroFlag=0 THEN PC := PC + y; END;
    $11:SetDE(FetchWord);
    $12:PutByte(DE, Regs[RegA]);
    $13:SetDE(DE+1);
    $14:Inc8(Regs[2]);
    $15:Dec8(Regs[2]);
    $16:Regs[2] := Fetch;
    $17:DoShiftOp(2, Regs[RegA]);
    $18:BEGIN y:=Fetch; PC := PC + y; END;
    $19:AddHL(DE);
    $1A:Regs[RegA] := ReadWord(DE);
    $1B:SetDE(DE-1);
    $1C:Inc8(Regs[3]);
    $1D:Dec8(Regs[3]);
    $1E:Regs[3] := Fetch;
    $1F:DoShiftOp(3, Regs[RegA]);
    $20:BEGIN y:=Fetch; IF Flags AND ZeroFlag=0 THEN PC := PC + y; END;
    $21:SetHL(FetchWord);
    $22:PutWord(FetchWord, HL);
    $23:SetHL(HL+1);
    $24:Inc8(Regs[4]);
    $25:Dec8(Regs[4]);
    $26:Regs[4] := Fetch;
    $27:DecimalAdjust;
    $28:BEGIN y:=Fetch; IF Flags AND ZeroFlag<>0 THEN PC := PC + y; END;
    $29:AddHL(HL);
    $2A:SetHL(ReadWord(FetchWord));
    $2B:SetHL(HL-1);
    $2C:Inc8(Regs[5]);
    $2D:Dec8(Regs[5]);
    $2E:Regs[5] := Fetch;
    $2F:Regs[RegA] := NOT Regs[RegA];
    $30:BEGIN y:=Fetch; IF Flags AND CarryFlag=0 THEN PC := PC + y; END;
    $31:SP := FetchWord;
    $32:PutByte(FetchWord, Regs[RegA]);
    $33:Inc(SP);
    $34:Inc8(Regs[6]);
    $35:Dec8(Regs[6]);
    $36:Regs[6] := Fetch;
    $37:Flags:=Flags OR CarryFlag;
    $38:BEGIN y:=Fetch; IF Flags AND CarryFlag<>0 THEN PC := PC + y; END;
    $39:AddHL(SP);
    $3A:Regs[RegA] := ReadWord(FetchWord);
    $3B:Dec(SP);
    $3C:Inc8(Regs[7]);
    $3D:Dec8(Regs[7]);
    $3E:Regs[7] := Fetch;
    $3F:Flags := Flags XOR CarryFlag;
  ELSE
    MonLog('Bad opcode ' + HexByte(opc), $70);
    broken:=TRUE;
  END;
END;

{---------------------------------------------------------------------------}
{ Opcodes 01??????b                                                         }
{---------------------------------------------------------------------------}
PROCEDURE Handle4(opc:BYTE);
BEGIN
  IF opc=$76 THEN BEGIN
    MonLog('HALT instruction',$1E);
    broken:=TRUE;
  END ELSE
   Regs[(opc SHR 3) AND 7] := Regs[opc AND 7];
END;

{---------------------------------------------------------------------------}
{ Opcodes 10??????b                                                         }
{---------------------------------------------------------------------------}
PROCEDURE Handle8(opc:BYTE);
BEGIN
  Arit((opc SHR 3) AND 7, Regs[opc AND 7]);
END;

{---------------------------------------------------------------------------}
{ Shift-Operationen (CBh + subcode)                                         }
{---------------------------------------------------------------------------}
PROCEDURE ShiftOp;
VAR opc:BYTE;
    tst:BYTE;
BEGIN
  opc:=Fetch;
  tst:=(opc SHR 3) AND 7;
  CASE opc AND $C0 OF
    $00:DoShiftOp(tst, Regs[opc AND 7]);
    $40:BEGIN
      Flags := (Flags AND NOT (NegativFlag OR Zeroflag)) OR HalfFlag;
      IF Regs[opc AND 7] AND (1 SHL tst)=0 THEN Flags:=Flags OR Zeroflag;
    END;
    $80:Regs[opc AND 7] := Regs[opc AND 7] AND NOT (1 SHL tst);
    $C0:Regs[opc AND 7] := Regs[opc AND 7] OR (1 SHL tst);
  END;
END;

{---------------------------------------------------------------------------}
{ Z80 Opcodes (EDh + subcode)                                               }
{---------------------------------------------------------------------------}
PROCEDURE Extension;
VAR opc:BYTE;

  PROCEDURE LDI;
  BEGIN
    PutByte(DE, ReadByte(HL));
    SetDE(DE+1);
    SetHL(HL+1);
    SetBC(BC-1);
    Flags := Flags AND NOT (NegativFlag + HalfFlag);
    SetFlag(PVflag, BC<>0);
  END;

  PROCEDURE LDD;
  BEGIN
    PutByte(DE, ReadByte(HL));
    SetDE(DE-1);
    SetHL(HL-1);
    SetBC(BC-1);
    Flags := Flags AND NOT (NegativFlag + HalfFlag);
    SetFlag(PVflag, BC<>0);
  END;

  PROCEDURE CPI;
  BEGIN
    Arit(7, ReadByte(HL));
  { IF Flags AND Zeroflag<>0 THEN Exit;     ??????????}
    SetHL(HL+1);
    SetBC(BC-1);
    SetFlag(PVflag, BC<>0);
  END;

  PROCEDURE CPD;
  BEGIN
    Arit(7, ReadByte(HL));
  { IF Flags AND Zeroflag<>0 THEN Exit;     ??????????}
    SetHL(HL-1);
    SetBC(BC-1);
    SetFlag(PVflag, BC<>0);
  END;

BEGIN
  opc:=Fetch;
  CASE opc OF
    $40:Regs[0] := InByte(Regs[RegC]);
    $41:OutByte(Regs[RegC], Regs[0]);
    $42:SbbHL(BC);
    $43:PutWord(FetchWord, BC);
    $44:Regs[RegA] := -Regs[RegA];
    $48:Regs[1] := InByte(Regs[RegC]);
    $49:OutByte(Regs[RegC], Regs[1]);
    $4A:AdcHL(BC);
    $4B:SetBC(ReadWord(FetchWord));
    $50:Regs[2] := InByte(Regs[RegC]);
    $51:OutByte(Regs[RegC], Regs[2]);
    $52:SbbHL(DE);
    $53:PutWord(FetchWord, DE);
   {FIXME: richtiges Handling fr I-Register}
    $57:BEGIN
      Regs[RegA] := $FF;
      SetFlag(PVflag, iff);
      Flags:=Flags AND NOT (NegativFlag + HalfFlag);
      SetSZ(Regs[RegA]);
    END;
    $58:Regs[3] := InByte(Regs[RegC]);
    $59:OutByte(Regs[RegC], Regs[3]);
    $5A:AdcHL(BC);
    $5B:SetDE(ReadWord(FetchWord));
    $60:Regs[4] := InByte(Regs[RegC]);
    $61:OutByte(Regs[RegC], Regs[4]);
    $62:SbbHL(HL);
    $68:Regs[5] := InByte(Regs[RegC]);
    $69:OutByte(Regs[RegC], Regs[5]);
    $6A:AdcHL(HL);
    $70:Regs[6] := InByte(Regs[RegC]);
    $71:OutByte(Regs[RegC], Regs[6]);
    $72:SbbHL(SP);
    $73:PutWord(FetchWord, SP);
    $78:Regs[7] := InByte(Regs[RegC]);
    $79:OutByte(Regs[RegC], Regs[7]);
    $7A:AdcHL(SP);
    $7B:SP := ReadWord(FetchWord);
    $A0:LDI;
    $A1:CPI;
    $A8:LDD;
    $A9:CPD;
    $B0:IF BC<>0 THEN BEGIN LDI; Dec(PC,2); END;
    $B1:IF BC<>0 THEN BEGIN CPI; IF Flags AND Zeroflag=0 THEN Dec(PC,2); END;
    $B8:IF BC<>0 THEN BEGIN LDD; Dec(PC,2); END;
    $B9:IF BC<>0 THEN BEGIN CPD; IF Flags AND Zeroflag=0 THEN Dec(PC,2); END;
    $FC:Inc(SymbolOffset, Pop);
    $FD:BiosCall;
    $FE:BEGIN
      MonLog('Program ends (BOOT/WBOOT).', $1E);
      Dec(PC,2);
      broken:=TRUE;
    END;
    $FF:Syscall;
  ELSE
    MonLog('Unsupported ED code ' + HexByte(opc), $70);
    broken:=TRUE;
  END;
END;

PROCEDURE HandleC(opc:BYTE); FORWARD;

{---------------------------------------------------------------------------}
{ Befehle mit Indexregistern DD/FD + subcode                                }
{---------------------------------------------------------------------------}
PROCEDURE Index(VAR x:WORD);
VAR opc:BYTE;
    w:WORD;
    y:BYTE;
BEGIN
  opc:=Fetch;
  CASE opc OF
    $09: AddDbl(x, BC);
    $19: AddDbl(x, DE);
    $21: x:=FetchWord;
    $22: PutWord(FetchWord, x);
    $23: Inc(x);
    $29: AddDbl(x, x);
    $2A: x := ReadWord(FetchWord);
    $2B: Dec(x);
    $39: AddDbl(x, SP);
    $E1: x := Pop;
    $E3: BEGIN w:=x; x:=ReadWord(SP); PutWord(SP,w); END;
    $E5: Push(x);
    $E9: PC := x;
    $F9: SP := x;

    $34..$36, $46, $56, $66, $70..$75, $77, $CB,
    $86, $96, $A6, $B6, $8E, $9E, $AE, $BE,
    $4E, $5E, $6E, $7E:BEGIN
      w := x + FetchS;
      y := Regs[RegM];
      Regs[RegM] := ReadByte(w);
      CASE opc AND $C0 OF
        $00:Handle0(opc);
        $40:Handle4(opc);
        $80:Handle8(opc);
        $C0:HandleC(opc);
      END;
      PutByte(w, Regs[RegM]);
      Regs[RegM] := y;
    END;
  ELSE
    MonLog('Unsupported IX/IY opcode ' + HexByte(opc), $70);
    broken:=TRUE;
  END;
END;

{---------------------------------------------------------------------------}
{ Opcodes 11??????b                                                         }
{---------------------------------------------------------------------------}
PROCEDURE HandleC(opc:BYTE);
VAR w:WORD;
    i:BYTE;
BEGIN
  CASE opc OF
    $C0:IF Flags AND Zeroflag=0 THEN PC:=Pop;
    $C1:SetBC(Pop);
    $C2:IF Flags AND ZeroFlag=0 THEN PC := FetchWord ELSE FetchWord;
    $C3:PC := FetchWord;
    $C4:IF Flags AND ZeroFlag=0 THEN Call(FetchWord) ELSE FetchWord;
    $C5:Push(BC);
    $C6:Arit(0, Fetch);
    $C7:Call(0);
    $C8:IF Flags AND Zeroflag<>0 THEN PC:=Pop;
    $C9:PC:=Pop;
    $CA:IF Flags AND ZeroFlag<>0 THEN PC := FetchWord ELSE FetchWord;
    $CB:ShiftOp;
    $CC:IF Flags AND ZeroFlag<>0 THEN Call(FetchWord) ELSE FetchWord;
    $CD:Call(FetchWord);
    $CE:Arit(1, Fetch);
    $CF:Call(8);
    $D0:IF Flags AND Carryflag=0 THEN PC:=Pop;
    $D1:SetDE(Pop);
    $D2:IF Flags AND CarryFlag=0 THEN PC := FetchWord ELSE FetchWord;
    $D3:OutByte(Fetch, Regs[RegA]);
    $D4:IF Flags AND CarryFlag=0 THEN Call(FetchWord) ELSE FetchWord;
    $D5:Push(DE);
    $D6:Arit(2, Fetch);
    $D7:Call($10);
    $D8:IF Flags AND Carryflag<>0 THEN PC:=Pop;
    $D9:FOR w:=0 TO 5 DO BEGIN i:=Shadow[w]; Shadow[w] := Regs[w]; Regs[w]:=i; END;
    $DA:IF Flags AND CarryFlag<>0 THEN PC := FetchWord ELSE FetchWord;
    $DB:Regs[RegA] := InByte(Fetch);
    $DC:IF Flags AND CarryFlag<>0 THEN Call(FetchWord) ELSE FetchWord;
    $DD:Index(IX);
    $DE:Arit(3, Fetch);
    $DF:Call($18);
    $E0:IF Flags AND PVflag=0 THEN PC:=Pop;
    $E1:SetHL(Pop);
    $E2:IF Flags AND PVFlag=0 THEN PC := FetchWord ELSE FetchWord;
    $E3:BEGIN w:=ReadWord(SP); PutWord(SP, HL); SetHL(w); END;
    $E4:IF Flags AND PVFlag=0 THEN Call(FetchWord) ELSE FetchWord;
    $E5:Push(HL);
    $E6:Arit(4, Fetch);
    $E7:Call($20);
    $E8:IF Flags AND Zeroflag<>0 THEN PC:=Pop;
    $E9:PC := HL;
    $EA:IF Flags AND PVFlag<>0 THEN PC := FetchWord ELSE FetchWord;
    $EB:BEGIN w:=HL; SetHL(DE); SetDE(w); END;
    $EC:IF Flags AND PVFlag<>0 THEN Call(FetchWord) ELSE FetchWord;
    $ED:Extension;
    $EE:Arit(5, Fetch);
    $EF:Call($28);
    $F0:IF Flags AND Signflag=0 THEN PC:=Pop;
    $F1:SetAF(Pop);
    $F2:IF Flags AND SignFlag=0 THEN PC := FetchWord ELSE FetchWord;
    $F3:iff:=FALSE;
    $F4:IF Flags AND SignFlag=0 THEN Call(FetchWord) ELSE FetchWord;
    $F5:Push(AF);
    $F6:Arit(6, Fetch);
    $F7:Call($30);
    $F8:IF Flags AND Signflag<>0 THEN PC:=Pop;
    $F9:SP:=HL;
    $FA:IF Flags AND SignFlag<>0 THEN PC := FetchWord ELSE FetchWord;
    $FB:iff:=TRUE;
    $FC:IF Flags AND SignFlag<>0 THEN Call(FetchWord) ELSE FetchWord;
    $FD:Index(IY);
    $FE:Arit(7, Fetch);
    $FF:Call($38);
  ELSE
    MonLog('Bad opcode ' + HexByte(opc), $70);
    broken:=TRUE;
  END;
END;

{---------------------------------------------------------------------------}
{ Prft, ob Interrupts anliegen                                             }
{---------------------------------------------------------------------------}
PROCEDURE CheckInts;
VAR L:LONGINT;
BEGIN
  IF timer_enabled AND (pc_timer > timer_tick) THEN BEGIN
    timer_tick := pc_timer + timer_interval;
    Call(timer_call);
  END;
END;

{---------------------------------------------------------------------------}
{ Hauptschleife                                                             }
{---------------------------------------------------------------------------}
PROCEDURE Run;
VAR opc:BYTE;
    madr,rm:WORD;
BEGIN
  REPEAT
    InOutRes := 0;
    IF iff THEN CheckInts;
    madr:=HL;
    rm := ReadByte(madr);
    Regs[RegM] := rm;
    Monitor;
    opc := Fetch;
    CASE opc AND $C0 OF
      $00:Handle0(opc);
      $40:Handle4(opc);
      $80:Handle8(opc);
      $C0:HandleC(opc);
    END;
    IF rm<>Regs[RegM] THEN PutByte(madr, Regs[RegM]);
    IF Mem[$40:$17] AND 16<>0 THEN broken:=TRUE;
  UNTIL FALSE;
END;

{===========================================================================}
{                                                                           }
{   Initialisierung CP/M Emulation                                          }
{                                                                           }
{===========================================================================}
CONST BIOS : ARRAY[0..62] OF BYTE=(
                     $ED, $FF, 0,       { call 5 }
                     $ED, $FE, 0,       { BIOS Ruf 0 = Kaltstart = HALT }
                     $ED, $FE, 0,       { BIOS Ruf 1 = Warmstart = HALT }
                     $ED, $FD, 2,       { const }
                     $ED, $FD, 3,       { conin }
                     $ED, $FD, 4,       { conout }
                     $ED, $FD, 5,       { list }
                     $ED, $FD, 6,       { punch }
                     $ED, $FD, 7,       { reader }
                     $ED, $FD, 8,       { home }
                     $ED, $FD, 9,       { seldsk }
                     $ED, $FD,10,       { settrk }
                     $ED, $FD,11,       { setsec }
                     $ED, $FD,12,       { setdma }
                     $ED, $FD,13,       { read }
                     $ED, $FD,14,       { write }
                     $ED, $FD,15,       { listst }
                     $ED, $FD,16,       { sectran }
                     $ED, $FD,17,       { const }
                     $ED, $FD,18,       { readerst }
                     $ED, $FD,19        { punchst }
             );

CONST DPB : ARRAY[0..14] OF BYTE=(
                     18, 0,             { Sektoren/Spur }
                     5,                 { Block-Shift }
                     31,                { Block-Mask ==> 4K/Block }
                     1,                 { Extent-Mask }
                     255, 3,            { 1024 Blocks ==> 4M/Disk }
                     0, 4,              { 1024 Verzeichniseintrge }
                     255, 0,            { 8 Blocks mit Verzeichnis }
                     0, 0,              { Check-Sektoren = 0 (fixed disk) }
                     0, 0               { Reservierte OS-Sektoren }
             );

{---------------------------------------------------------------------------}
{ /num/ DPHs an Adresse /Adr/ aufbauen                                      }
{---------------------------------------------------------------------------}
PROCEDURE BuildDPHs(adr:WORD; num:INTEGER);
BEGIN
  WHILE num>0 DO BEGIN
    Dec(num);
    PutWord(adr, 0);            { keine Sektorbersetzung }
    PutWord(adr+2, 0);          { 6 Bytes reserviert }
    PutWord(adr+4, 0);
    PutWord(adr+6, 0);
    PutWord(adr+8, CPMscratch); { Verzeichnispuffer }
    PutWord(adr+10, CPMdpb);    { DPB-Zeiger }
    PutWord(adr+12, CPMscratch); { fr Diskettenwechsel-Test, den wir nicht machen }
    PutWord(adr+14, CPMalv);    { Allocation Vector (aka FAT) }
    Inc(adr, 16);
  END;
END;

{---------------------------------------------------------------------------}
{ Allocation Vector aufbauen                                                }
{---------------------------------------------------------------------------}
PROCEDURE BuildAlv(adr:WORD);
CONST AllocatedBlocksDIV8 = 20;
VAR i:INTEGER;
BEGIN
  PutWord(adr, ReadWord(CPMdpb + 9));   { Verzeichnis }
  FOR i:=1 TO AllocatedBlocksDIV8 DO PutByte(adr+1+i, 255);
END;

VAR CPMsize:WORD;
    call5:WORD;

BEGIN
  IF ParamCount=0 THEN BEGIN
    Writeln('Usage: ',ParamStr(0), ' image [mapfile|-]');
    Halt;
  END;
  InitCursor;
  monitoring := ParamCount>=2;
  IF monitoring THEN TextMode(co80 + font8x8) ELSE TextMode(co80);
  InitMonitor;
  GotoXY(1,25);
  TextAttr:=$70;
  ClrEOL;
  Window(1,1,80,24);
  TextAttr:=7;
  ClrScr;
  New(M);
  FillChar(M^, Sizeof(M^), 0);
  broken:=FALSE;
  brk := Sizeof(M^);

  sys := $FFFF;
  InitFilesys;
  LoadImage;

  CPMsize := 128 + 128 + CPMNumDrives*16 + Sizeof(DPB) + Sizeof(BIOS);

  IF sys = $FFFF THEN BEGIN
   { Suche Platz fr System }
    sys := brk - CPMsize;
    IF VRAM.Enabled THEN BEGIN
      IF (VRAM.At1 >= sys) AND (VRAM.At0 < sys) THEN sys := VRAM.At0 - CPMsize;
    END;
  END;
  IF (VRAM.Enabled) AND (sys > VRAM.At0) THEN call5 := VRAM.At0-3 ELSE call5 := sys;

  CPMdpb     := sys + Sizeof(BIOS);
  CPMdphbase := CPMdpb + Sizeof(DPB);
  CPMscratch := CPMdphbase + CPMNumDrives*16;
  CPMalv     := CPMscratch + 128;

  m^[0] := $C3;         { jmp 0 --> BIOS-Sprungvektor }
  m^[1] := Lo(sys+6);
  m^[2] := Hi(sys+6);
  m^[5] := $C3;         { eb ff    call 5 --> SYSCALL }
  m^[6] := Lo(call5);
  m^[7] := Hi(call5);
  FillChar(m^[$5D], 11, 32);    { FCBs }
  FillChar(m^[$6D], 11, 32);
  m^[$80] := 0;

  IF call5 <> sys THEN BEGIN
    m^[call5] := $C3;
    m^[call5+1] := Lo(sys);
    m^[call5+2] := Hi(sys);
  END;

  Move(BIOS, m^[sys], Sizeof(BIOS));
  Move(DPB, m^[CPMdpb], Sizeof(DPB));
  BuildDPHs(CPMdphbase, CPMNumDrives);
  BuildAlv(CPMalv);
  FillChar(Regs, Sizeof(Regs), 0);
  iff := TRUE;
  timer_enabled := FALSE;

  TextAttr:=11;
  Writeln('BICEM: ', call5-256, ' bytes TPA (0100-',HexWord(call5-1),'), ',
          CPMNumDrives, ' drives, BDOS at ',HexWord(sys));
  Writeln;
  TextAttr:=15;
  LoadInsns;
  TextAttr:=11;
  IF Monitoring THEN BEGIN
    Writeln('Monitor active. Press [r] to run program.');
  END;
  PC:=$100;
  SP:=sys-2;              { ret =^ jmp 0 }
  Run;
  DoHalt;
END.
