{---------------------------------------------------------------------------}
{                                                                           }
{  Dateizugriff fr BICEMU                                                  }
{                                                                           }
{  Jede Datei erhlt beim ffnen eine Nummer /free_index/ zugewiesen, die   }
{  in einem reservierten Feld des FCB gespeichert wird. Der FCB kann also   }
{  kopiert, verschoben etc werden, und wird immer noch erkannt.             }
{                                                                           }
{  Die Semantik von mehrfach geffneten und nicht geschlossenen Dateien     }
{  ist evtl. etwas anders als im original CP/M.                             }
{                                                                           }
{  (c) copyright 1999 Stefan Reuther <Streu@gmx.de>                         }
{                                                                           }
{---------------------------------------------------------------------------}
{$A+,B-,D+,E+,F-,G+,I-,L+,N-,O+,R-,S+,V-,X+}
{$M 16384,0,655360}
UNIT FileIO;

INTERFACE

TYPE PFile = ^TFile;
     TFile = RECORD
            Next:PFile;
            Drive:BYTE;
            Name:STRING[15];
            F:FILE;
            Size:LONGINT;
            index:WORD;         { `interner Systemgebrauch' }
            ref:INTEGER;
     END;

CONST CPMNumDrives = 10;        { max. 16 }

FUNCTION OpenFile(Drive:BYTE; Name:STRING; VAR PF:PFile; crea:BOOLEAN):BYTE;
FUNCTION ReadRecord(index,extent,rec:WORD; VAR buf):BYTE;
FUNCTION WriteRecord(index,extent,rec:WORD; VAR buf):BYTE;
FUNCTION CloseFile(index:WORD):BYTE;
PROCEDURE InitFileSys;
PROCEDURE MountDrive(Drive:BYTE; Path:STRING);
PROCEDURE UnmountDrive(Drive:BYTE);

VAR DrivePrefix   : ARRAY[0..CPMNumDrives-1] OF STRING[63];
    DrivesMounted : WORD;

IMPLEMENTATION

CONST openfiles:PFile = NIL;
      free_index:WORD = 0;

{---------------------------------------------------------------------------}
{ Datei /name/ auf LW /drive/ ffnen, oder erzeugen bei /crea/=true.        }
{ Liefert Code fr Register A, und den Dateirecord in /PF/                  }
{ /drive/: 0-based                                                          }
{---------------------------------------------------------------------------}
FUNCTION OpenFile(Drive:BYTE; Name:STRING; VAR PF:PFile; crea:BOOLEAN):BYTE;
VAR p:PFile;
BEGIN
  IF NOT Odd(DrivesMounted SHR Drive) THEN BEGIN
    OpenFile:=$FF;
    Exit;
  END;
  p:=openfiles;
  WHILE (p<>NIL) AND ((p^.Name<>Name) OR (p^.Drive<>Drive)) DO p:=p^.Next;
  IF p=NIL THEN BEGIN
    New(p);
    p^.Next:=NIL;
    p^.Name:=Name;
    Assign(p^.f, DrivePrefix[Drive] + Name);
    IF crea THEN Rewrite(p^.f, 1) ELSE Reset(p^.f, 1);
    IF IOResult<>0 THEN BEGIN
      Dispose(p);
      OpenFile:=$FF;
      Exit;
    END;
    p^.Index:=Free_Index;
    Inc(free_index);
    p^.Size:=FileSize(p^.F);
    p^.Ref:=0;
    p^.Drive:=Drive;
    p^.Next:=OpenFiles;
    OpenFiles:=p;
  END;
  Inc(p^.Ref);
  pf := p;
  OpenFile:=0;
END;

{---------------------------------------------------------------------------}
{ Datei mit Index /index/ raussuchen                                        }
{---------------------------------------------------------------------------}
FUNCTION Lookup(index:WORD):PFile;
VAR pf:PFile;
BEGIN
  pf:=OpenFiles;
  WHILE (pf<>NIL) AND (pf^.Index<>index) DO pf:=pf^.Next;
  Lookup:=pf;
END;

{---------------------------------------------------------------------------}
{ Record (128 bytes) an spezifiziertem /extent/, /rec/ lesen, /index/ =     }
{ reserviertes Feld aus dem FCB                                             }
{---------------------------------------------------------------------------}
FUNCTION ReadRecord(index,extent,rec:WORD; VAR buf):BYTE;
VAR pf:PFile;
    L:LONGINT;
    R:WORD;
BEGIN
  pf:=Lookup(index);
  IF pf=NIL THEN ReadRecord:=$FF ELSE BEGIN
    L:=(rec + LONGINT(128)*extent) * 128;
    IF L>=pf^.Size THEN ReadRecord:=1 ELSE BEGIN
      Seek(pf^.F, L);
      FillChar(buf, 128, 0);
      Blockread(pf^.F, buf, 128, R);
      ReadRecord:=0;
    END;
  END;
END;

{---------------------------------------------------------------------------}
{ Record (128 bytes) an spezifiziertem /extent/, /rec/ schreiben, /index/ = }
{ reserviertes Feld aus dem FCB                                             }
{---------------------------------------------------------------------------}
FUNCTION WriteRecord(index,extent,rec:WORD; VAR buf):BYTE;
VAR pf:PFile;
    L:LONGINT;
    R:WORD;
BEGIN
  pf:=Lookup(index);
  IF pf=NIL THEN WriteRecord:=$FF ELSE BEGIN
    L:=(rec + LONGINT(128)*extent) * 128;
    Seek(pf^.F, L);
    BlockWrite(pf^.F, buf, 128, R);
    Inc(L, R);
    IF pf^.Size<L THEN pf^.Size:=L;
    IF R<>128 THEN WriteRecord:=1 ELSE WriteRecord:=0;
  END;
END;

{---------------------------------------------------------------------------}
{ Datei /index/ schlieen                                                   }
{---------------------------------------------------------------------------}
FUNCTION CloseFile(index:WORD):BYTE;
VAR pf,pf1:PFile;
BEGIN
  pf:=Lookup(index);
  IF pf=NIL THEN CloseFile:=$FF ELSE BEGIN
    Dec(pf^.Ref);
    IF pf^.Ref=0 THEN BEGIN
      Close(pf^.F);
      IF pf=OpenFiles THEN OpenFiles:=pf^.Next ELSE BEGIN
        pf1:=OpenFiles;
        WHILE pf1^.Next<>pf DO pf1:=pf1^.Next;
        pf1^.Next:=pf^.Next;
      END;
      Dispose(pf);
    END;
    CloseFile:=0;
  END;
END;

{---------------------------------------------------------------------------}
{ Initialisierung Dateisystem                                               }
{---------------------------------------------------------------------------}
PROCEDURE InitFileSys;
VAR i:INTEGER;
BEGIN
  FOR i:=0 TO CPMNumDrives-1 DO DrivePrefix[i]:='.\';
  DrivesMounted := 3; { a,b gemountet }
END;

{---------------------------------------------------------------------------}
{ Laufwerk /drive/ mit Pfad /path/ auf dem Hostsystem verbinden             }
{---------------------------------------------------------------------------}
PROCEDURE MountDrive(Drive:BYTE; Path:STRING);
BEGIN
  UnmountDrive(Drive);
  DrivesMounted := DrivesMounted OR (1 SHL Drive);
  IF (Path<>'') AND (Path[Length(Path)]<>'\') AND (Path[Length(Path)]<>':') THEN
   Path:=Path+'\';
  DrivePrefix[Drive] := Path;
END;

{---------------------------------------------------------------------------}
{ Laufwerk /drive/ abhngen                                                 }
{ schliet alle Dateien auf diesem LW                                       }
{---------------------------------------------------------------------------}
PROCEDURE UnmountDrive(Drive:BYTE);
VAR pf,pf1,pfn:PFile;
BEGIN
  IF NOT Odd(DrivesMounted SHR Drive) THEN Exit;
  pf1:=NIL;
  pf:=OpenFiles;
  WHILE pf<>NIL DO BEGIN
    pfn:=pf^.Next;
    IF pf^.Drive = Drive THEN BEGIN
      Close(pf^.F);
      IF pf1=NIL THEN OpenFiles:=pfn ELSE pf1^.Next:=pfn;
      Dispose(pf);
    END ELSE BEGIN
      pf1:=pf;
    END;
    pf:=pfn;
  END;
  DrivesMounted:=DrivesMounted AND NOT (1 SHL Drive);
  IF Drive=0 THEN BEGIN
    DrivesMounted:=DrivesMounted OR 1;
    DrivePrefix[0] := '.\';
  END;
END;

END.
