;
;  BICEM - Geschwindigkeitsmessung
;
;  Verfahren: es wird eine `Endlosschleife' gefahren, die von einem
;  Timerinterrupt unterbrochen werden kann. Diese Unterbrechung
;  ist derart realisiert, da der Timer nicht direkt an die unter-
;  brochene Code-Stelle zurckkehrt, sondern eine Adresse vom Stack
;  poppt.
;
;  (c) copyright 1999 Stefan Reuther
;

                ld      de,Starttext
                ld      c,9
                call    5

                call    GetTimer                ; Alten Timer sichern
                push    de
                push    bc
                push    af

                ld      hl,LeaveLoop            ; Stack fr Zeitmessung
                push    hl
                ld      hl,EnterLoop
                push    hl
                ld      hl,EnterLoop1
                push    hl

                di
                ld      de,Timer                ; 18 Ticks abwarten
                ld      bc,18                   ; -> ~1 / Sekunde
                call    SetTimer

                ld      hl,0                    ; Initialisierung Zhlschleife
                ld      b,0

; Erste Schleife: warten auf einen Timer-Interrupt
Eieiei:         ei
                jr      Eieiei

; Zweite Schleife: ein Timer ist evtl. verzgert (wegen di) -> noch
; einen Timer abwarten
EnterLoop1:     ei
                jr      EnterLoop1

; Haupt-Zhlschleife: mit hoher Wahrscheinlichkeit nutzt diese
; Schleife eine komplette Timer-Periode.
EnterLoop:      ei
                inc     b
                jrnz    EnterLoop
                inc     hl
                jr      EnterLoop

; Alten Timer restaurieren
LeaveLoop:      pop     af
                pop     bc
                pop     de
                canc    StopTimer
                cac     SetTimer
                ei

                push    hl              ; Ergebnis ausgeben
                call    PrintDecNumber
                ld      de,fpstext
                ld      c,9
                call    5
                rst     0

;---------------------------------------------------------------------------
GetTimer:       db      $ED,$FD,130
SetTimer:       db      $ED,$FD,128
StopTimer:      db      $ED,$FD,129

;---------------------------------------------------------------------------
; Interrupt. Mu di setzen: wenn ein Timer zuviel ausgefhrt wird -> Absturz
;---------------------------------------------------------------------------
Timer:          di
                pop     de
                ret

;---------------------------------------------------------------------------
; Zahl auf dem Stack dezimal ausgeben; nicht reentrant
;---------------------------------------------------------------------------
PrintDecNumber: pop     hl
                ex      (sp),hl
                ld      ix,DigitBuffer          ; als fnfstellige Zahl
                ld      bc,9999                 ; formatieren
                call    Digit
                ld      bc,999
                call    Digit
                ld      bc,99
                call    Digit
                ld      c,9
                call    Digit
                ld      (ix),l

                ld      ix,DigitBuffer          ; maximal 4 fhrende Nullen
                ld      b,4                     ; entfernen
  SkipLoop:     ld      a,(ix)
                cmp     0
                jrnz    PrintLoop
                inc     ix
                djnz    SkipLoop
  PrintLoop:    ld      a,(ix)                  ; Ausgabe
                add     $30
                rc                              ; Carry entsteht, wenn der
                ld      e,a                     ; Sentinel erreicht wird
                ld      c,2
                call    5
                inc     ix
                jr      PrintLoop

;--- Unterprogramm: *ix++ = HL/(BC+1), HL %= (BC+1) durch Subtraktion ---
Digit:          ld      a,-1
  DigitLoop:    inc     a
                scf
                sbc     hl,bc
                jrnc    DigitLoop
                adc     hl,bc
                ld      (ix),a
                inc     ix
                ret

DigitBuffer:    space   5
                db      $FF     ; Sentinel

;=========================================================================
Starttext:      db      'Speed test for BICEM',13,10,10,36
fpstext:        db      ' iterations per second.',13,10,36
