{===========================================================================}
{                                                                           }
{    Z80 Assembler                                                          }
{                                                                           }
{===========================================================================}
{                                                                           }
{   Dieser Assembler basiert auf derselben Steuerdatei wie das Z80decode    }
{   Modul. Instruktionen werden per pattern matching auf die Eingabedatei   }
{   gematcht. Damit ist der Assembler langsam, aber relativ kurz. Der       }
{   Code ist auf 10 kByte begrenzt.                                         }
{                                                                           }
{   (c) copyright 1999 Stefan Reuther                                       }
{                                                                           }
{===========================================================================}
USES StrUtil;

TYPE TDblRegs = ARRAY[0..3] OF STRING[2];

{=== Symbolnamen ===}
TYPE PMap = ^TMap;
     TMap = RECORD
         Next:PMap;
         Name:STRING[63];
         Ofs:WORD;
     END;

CONST Map:PMap = NIL;

{--- reservierte Symbolnamen ---}
CONST Reserved = 'B C D E H L BC DE HL SP AF IX IY';

CONST RegNames:ARRAY[0..7] OF STRING[4] = ('B', 'C', 'D', 'E', 'H', 'L', '(HL)', 'A');
      ddRegs:TDblRegs = ('BC','DE','HL','SP');
      ppRegs:TDblRegs = ('BC','DE','??','SP');
      qqRegs:TDblRegs = ('BC','DE','HL','AF');

{=== Befehlslisting ===}
TYPE PInsn = ^TInsn;
     TInsn = RECORD
         Next:PInsn;
         Mask,Res:BYTE;
         Status:LONGINT;        { bit0 = initial, bit1 = 'a', ... }
         Name:STRING[7];
         Action:STRING[31];
     END;

CONST Insns : PInsn = NIL;

{---------------------------------------------------------------------------}
{ Statusbits aus Zeichenfolge berechnen                                     }
{ @ = bit 0                                                                 }
{ A = bit 1    bis                                                          }
{ Z = bit 26                                                                }
{---------------------------------------------------------------------------}
FUNCTION NewStatus(S:STRING):LONGINT;
VAR L:LONGINT;
    c:CHAR;
BEGIN
  L:=0;
  REPEAT
    IF (S='') OR (S[1]=']') THEN BEGIN
      NewStatus:=L;
      Exit;
    END;
    c := Upcase(S[1]);
    Delete(S,1,1);
    IF (c>='@') AND (c<='Z') THEN L:=L OR (LONGINT(1) SHL (Ord(c)-Ord('@')));
  UNTIL FALSE;
END;

{---------------------------------------------------------------------------}
{ INSN.LST laden                                                            }
{---------------------------------------------------------------------------}
PROCEDURE LoadInsns;
VAR F:TEXT;
    S:STRING;
    Status:LONGINT;
    i,j,k:INTEGER;
    this:PInsn;
BEGIN
  S:=ParamStr(0);
  WHILE (S<>'') AND (S[Length(S)]<>':') AND (S[Length(S)]<>'\') DO Dec(S[0]);
  Assign(F, S+'insn.lst');
 {$I-}
  Reset(F);
 {$I+}
  IF IOresult<>0 THEN BEGIN
    Writeln('Could not load assembler data: exit.');
    Halt;
  END;
  Status := 1;
  WHILE NOT EOF(F) DO BEGIN
    Readln(F,S);
    IF (S<>'') AND (S[1]<>';') THEN BEGIN
      IF S[1]='[' THEN BEGIN
        Status:=NewStatus(S);
      END ELSE BEGIN
        this := New(PInsn);
        this^.Status := Status;
        this^.Mask := 0;
        this^.Res := 0;
        FOR i:=1 TO 8 DO BEGIN
          IF S='' THEN BEGIN
            Writeln('Error in insn definition file');
            Halt;
          END;
          this^.Res := this^.Res*2;
          this^.Mask := this^.Mask*2;
          IF S[1]='1' THEN Inc(this^.Res);
          IF S[1]<>'.' THEN Inc(this^.Mask);
          Delete(S,1,1);
        END;
        Strip(S,s_Both);
        i:=Pos(' ',S);
        IF i>0 THEN BEGIN
          this^.Name:=UpStr(Copy(S,1,i));
          this^.Action:=Copy(S,i,255) + ';';
        END ELSE BEGIN
          this^.Name:=Upstr(S);
          this^.Action:=';';
        END;
        Strip(this^.Name, s_Both);
        Strip(this^.Action, s_Both);
        this^.Next:=insns;
        insns:=this;
      END;
    END;
  END;
END;

{=== Argumente fr Befehle ===}
VAR mr:BYTE;            { %r/%b }
    mrr:BYTE;           { %R }
    mv:BYTE;            { %v }
    mimm:STRING;        { Rest (%i,%w,%j) }
    md:BYTE;            { %d,%q,%p }
    mx:CHAR;            { %x }
    margs:STRING[20];   { Liste der verwendeten Kommandozeichen }

CONST idchars = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_+-$';

{---------------------------------------------------------------------------}
{ TRUE, wenn /S/ eins der in /r/ angegebenen Doppelregister ist             }
{ in diesem Fall wird /md/ entsprechend gesetzt                             }
{---------------------------------------------------------------------------}
FUNCTION MatchDblReg(VAR r:TDblRegs; S:STRING):BOOLEAN;
VAR i:INTEGER;
BEGIN
  FOR i:=0 TO 3 DO IF r[i]=S THEN BEGIN
    md:=i;
    MatchDblReg:=TRUE;
    Exit;
  END;
  MatchDblReg:=FALSE;
END;

{---------------------------------------------------------------------------}
{ Matcht Argumentliste /S/ auf Instruktion /p/'s Action                     }
{ /S/ enthlt keine Spaces mehr, und in Grobuchstaben                      }
{---------------------------------------------------------------------------}
FUNCTION MatchInsn(p:PInsn; S:STRING):BOOLEAN;
VAR aptr,sptr:INTEGER;
    i,j:INTEGER;
LABEL Break;
BEGIN
  margs:='';
  mx:=' ';
  aptr:=1;
  sptr:=1;
  REPEAT
    IF p^.Action[aptr]='%' THEN BEGIN
      Inc(aptr);
      CASE p^.Action[aptr] OF
       {=== Register ===}
        'r','R':BEGIN
          FOR i:=0 TO 7 DO IF Copy(S,sptr,Length(RegNames[i]))=RegNames[i] THEN BEGIN
            Inc(sptr, Length(RegNames[i])-1);
            IF p^.Action[aptr]='r' THEN mr:=i ELSE mrr:=i;
            margs:=margs + p^.Action[aptr];
            GOTO Break;
          END;
        END;
       {=== Zahlenwerte ===}
        'j','w','i':BEGIN
          margs:=margs + p^.Action[aptr];
          mimm:='';
          WHILE (Pos(S[sptr],idchars)<>0) AND ((aptr=Length(p^.Action)) OR (S[sptr]<>p^.Action[aptr])) DO BEGIN
            mimm:=mimm + S[sptr];
            Inc(sptr);
          END;
          IF Pos(Copy(mimm,1,2),Reserved)<>0 THEN BEGIN
            MatchInsn:=FALSE;
            Exit;
          END;
          Dec(sptr);
        END;
        'v':BEGIN
          margs:=margs + p^.Action[aptr];
          i:=0;
          IF S[sptr]='-' THEN BEGIN
            Inc(sptr);
            i:=1;
          END ELSE IF S[sptr]='+' THEN Inc(sptr);
          mv:=0;
          WHILE (S[sptr]>='0') AND (S[sptr]<='9') DO BEGIN
            mv := 10*mv + Ord(S[sptr])-Ord('0');
            Inc(sptr);
          END;
          Dec(sptr);
          IF i<>0 THEN mv:=-mv;
        END;
       {=== Doppelregister ===}
        'd':IF MatchDblReg(ddRegs, Copy(S,sptr,2)) THEN BEGIN
          Inc(sptr);
          margs:=margs+p^.Action[aptr];
        END ELSE BEGIN MatchInsn:=FALSE; Exit; END;
        'p':IF MatchDblReg(ppRegs, Copy(S,sptr,2)) THEN BEGIN
          Inc(sptr);
          margs:=margs+p^.Action[aptr];
        END ELSE BEGIN MatchInsn:=FALSE; Exit; END;
        'q':IF MatchDblReg(qqRegs, Copy(S,sptr,2)) THEN BEGIN
          Inc(sptr);
          margs:=margs+p^.Action[aptr];
        END ELSE BEGIN MatchInsn:=FALSE; Exit; END;
       {=== Status ===}
        'x':BEGIN
          margs:=margs+p^.Action[aptr];
          mx:=S[sptr];
        END;
       {=== Bitnummer (Ziffer) ===}
        'b':IF (S[sptr]>='0') AND (S[sptr]<='7') THEN mr:=Ord(S[sptr]) - 48
        ELSE BEGIN
          MatchInsn:=FALSE;
          Exit;
        END;
      END;
     Break:
    END ELSE
    IF p^.Action[aptr]=';' THEN BEGIN
      MatchInsn:=S[sptr]=';';
      Exit;
    END ELSE
    IF Upcase(p^.Action[aptr])<>S[sptr] THEN BEGIN
      MatchInsn:=FALSE;
      Exit;
    END;
    Inc(aptr);
    Inc(sptr);
  UNTIL FALSE;
END;

{---------------------------------------------------------------------------}
{ Adresse auf Name abbilden                                                 }
{---------------------------------------------------------------------------}
FUNCTION Mapper(S:STRING):PMap;
VAR p:PMap;
BEGIN
  p:=Map;
  WHILE (p<>NIL) AND (p^.Name<>S) DO p:=P^.Next;
  Mapper:=p;
END;

{---------------------------------------------------------------------------}
{ Symbol erzeugen                                                           }
{---------------------------------------------------------------------------}
PROCEDURE AddSymbol(S:STRING; ip:WORD);
VAR p:PMap;
BEGIN
  S:=Upstr(S);
  IF Mapper(S)<>NIL THEN BEGIN
    Writeln('Duplicate identifier ',S);
  END ELSE BEGIN
    New(p);
    p^.ofs := IP;
    p^.Name := S;
    p^.Next := Map;
    Map := p;
  END;
END;

{---------------------------------------------------------------------------}
{ Assembler                                                                 }
{---------------------------------------------------------------------------}
VAR ip:WORD;
    buffer:ARRAY[1..10240] OF BYTE;

PROCEDURE wb(y:BYTE);
BEGIN
  Inc(ip);
  buffer[ip-$100] := y;
END;

{---------------------------------------------------------------------------}
{ Befehl /op,args/ identifizieren                                           }
{---------------------------------------------------------------------------}
FUNCTION Identify(op,args:STRING):PInsn;
VAR P:PInsn;
    i:INTEGER;
BEGIN
  FOR i:=Length(args) DOWNTO 1 DO IF args[i]=' ' THEN Delete(args,i,1)
  ELSE args[i]:=Upcase(args[i]);
  args:=args+';';
  p:=Insns;
  WHILE p<>NIL DO BEGIN
    IF (op=p^.Name) AND MatchInsn(p,args) THEN BEGIN
      Identify:=p;
      Exit;
    END;
    P:=p^.Next;
  END;
  Identify:=NIL;
END;

{---------------------------------------------------------------------------}
{ true, wenn /b/ ein Zeichen aus /a/ enthlt                                }
{---------------------------------------------------------------------------}
FUNCTION PosN(a,b:STRING):BOOLEAN;
VAR i:INTEGER;
BEGIN
  FOR i:=1 TO Length(a) DO IF Pos(a[i],b)<>0 THEN BEGIN
    PosN:=TRUE;
    Exit;
  END;
  PosN:=FALSE;
END;

{---------------------------------------------------------------------------}
{ Zahl oder Bezeichner /s/ auswerten                                        }
{---------------------------------------------------------------------------}
FUNCTION ParseThing(S:STRING):INTEGER;
VAR i,j:INTEGER;
    p:PMap;
BEGIN
  IF S='' THEN BEGIN
    ParseThing:=0;
    Exit;
  END;
  Val(S,i,j);
  IF j=0 THEN ParseThing:=i ELSE BEGIN
    p:=Mapper(S);
    IF p=NIL THEN BEGIN
      Writeln('Unknown identifier ',S);
      ParseThing:=0;
    END ELSE ParseThing:=p^.ofs;
  END;
END;

{---------------------------------------------------------------------------}
{ Argument /s/ (Summe aus Zahlen/Ids) auswerten                             }
{---------------------------------------------------------------------------}
FUNCTION ParseArg(S:STRING):INTEGER;
VAR S1:STRING;
    total:WORD;
    op:(plus,minus);
BEGIN
  Strip(S, s_Both);
  total:=0;
  op:=plus;
  S:=S+'+';
  S1:='';
  WHILE S<>'' DO BEGIN
    IF S[1]=';' THEN S:='' ELSE
    IF (S[1]='+') OR (S[1]='-') THEN BEGIN
      IF op=plus THEN Inc(total,ParseThing(S1)) ELSE Dec(total, ParseThing(S1));
      IF S[1]='+' THEN op:=plus ELSE op:=minus;
      S1:='';
    END ELSE S1:=S1+S[1];
    Delete(S,1,1);
  END;
  ParseArg:=total;
END;

{---------------------------------------------------------------------------}
{ 'db'-Befehl                                                               }
{---------------------------------------------------------------------------}
PROCEDURE HandleDB(S:STRING; Pass:INTEGER);
VAR i,j:INTEGER;
    c:CHAR;
BEGIN
  WHILE S<>'' DO BEGIN
    Strip(S,s_Both);
    IF S[1]=';' THEN Exit ELSE
    IF (S[1]='''') OR (S[1]='"') THEN BEGIN
      c:=S[1];
      Delete(S,1,1);
      i:=Pos(c,S);
      IF i=0 THEN i:=Length(S)+1;
      FOR j:=1 TO i-1 DO wb(Ord(S[j]));
      Delete(S,1,i);
    END ELSE IF S[1]=',' THEN Delete(S,1,1) ELSE
    BEGIN
      i:=Pos(',',S);
      j:=Pos(';',S);
      IF (i=0) OR ((j<>0) AND (j<i)) THEN i:=j;
      IF i=0 THEN i:=Length(S) ELSE Dec(i);
      IF Pass=1 THEN wb(0) ELSE
       WB(ParseArg(Upstr(Copy(S,1,i))));
      Delete(S,1,i);
    END;
  END;
END;

{---------------------------------------------------------------------------}
{ 'dw'-Befehl                                                               }
{---------------------------------------------------------------------------}
PROCEDURE HandleDW(S:STRING; Pass:INTEGER);
VAR i,j:INTEGER;
BEGIN
  WHILE S<>'' DO BEGIN
    Strip(S,s_Both);
    IF S[1]=';' THEN Exit ELSE
    IF S[1]=',' THEN Delete(S,1,1) ELSE
    BEGIN
      i:=Pos(',',S);
      IF i=0 THEN i:=Length(S)+1;
      IF Pass=1 THEN BEGIN
        wb(0); wb(0);
      END ELSE BEGIN
        j := ParseArg(Upstr(Copy(S,1,i-1)));
        wb(j);
        wb(j SHR 8);
      END;
      Delete(S,1,i);
    END;
  END;
END;

{---------------------------------------------------------------------------}
{ 'space'-Befehl                                                            }
{---------------------------------------------------------------------------}
PROCEDURE HandleSPACE(S:STRING);
VAR i,j:INTEGER;
BEGIN
  i:=Pos(';',S);
  IF i>0 THEN S[0] := Chr(i-1);
  Strip(S, s_both);
  Val(S, i, j);
  IF j<>0 THEN BEGIN
    Writeln('SPACE must have a constant argument.');
    Exit;
  END ELSE BEGIN
    FOR j:=1 TO i DO WB(0);
  END;
END;

{---------------------------------------------------------------------------}
{ Suche Weg von Status /src/ in /dest/, /vissrc/ = Statuscode fr src       }
{ (Zeichen), /visdest/ zu erreichender Statuscode, ' ' wenn egal            }
{---------------------------------------------------------------------------}
FUNCTION BacktrackStatus(src, dest:LONGINT; vissrc,visdest:CHAR; VAR res:STRING):BOOLEAN;
VAR p:PInsn;
    new:LONGINT;
    newvs:CHAR;
    i:INTEGER;
BEGIN
  IF (src AND dest<>0) AND ((visdest=' ') OR (visdest=vissrc)) THEN BEGIN
    res := '';
    BacktrackStatus := TRUE;
    Exit;
  END;
  p := Insns;
  WHILE p<>NIL DO BEGIN
    IF (p^.Status AND src<>0) AND (p^.Name[1]='>') THEN BEGIN
     {=== Statusbergang von src ===}
      IF p^.Name[2]='>' THEN i:=3 ELSE i:=2;
      new := LONGINT(1) SHL (Ord(p^.Name[i]) - Ord('@'));
      IF i=3 THEN newvs:=vissrc ELSE newvs:=p^.Name[2];
      IF BacktrackStatus(new,dest,newvs,visdest,res) THEN BEGIN
        res:=Chr(p^.Res) + res;
        BacktrackStatus:=TRUE;
        Exit;
      END;
    END;
    p:=p^.Next;
  END;
  BacktrackStatus:=FALSE;
END;

{---------------------------------------------------------------------------}
{ Befehl 'p' assemblieren                                                   }
{---------------------------------------------------------------------------}
PROCEDURE AssembleInsn(p:PInsn; pass:INTEGER);
VAR arg,i:INTEGER;
    opc:BYTE;
    s:STRING;
BEGIN
  IF p^.Status<>1 THEN BEGIN
    IF NOT BacktrackStatus(1, p^.Status, '@', mx, s) THEN BEGIN
      Writeln('Can''t find status transitions for ',p^.Name,' ',p^.Action);
      Exit;
    END;
    FOR i:=1 TO Length(S) DO wb(Ord(S[i]));
  END;
  opc:=p^.Res;
  IF PosN('rb',margs) THEN opc:=opc OR (mr SHL 3);
  IF Pos('R',margs)<>0 THEN opc:=opc OR mrr;
  IF PosN('dpq',margs) THEN opc:=opc OR (md SHL 4);
  wb(opc);
  IF Pos('v',margs)<>0 THEN wb(mv);
  IF PosN('jwi',margs) THEN BEGIN
    IF Pass=1 THEN BEGIN
      wb(0);
      IF Pos('w',margs)<>0 THEN wb(0);
    END ELSE BEGIN
      arg := ParseArg(mimm);
      IF Pos('i',margs)<>0 THEN wb(arg) ELSE
      IF Pos('w',margs)<>0 THEN BEGIN
        wb(arg);
        wb(arg SHR 8);
      END ELSE
      IF Pos('j',margs)<>0 THEN BEGIN
        arg := arg - (IP+1);
        IF (arg>127) OR (arg<-128) THEN Writeln('Relative jump out of range');
        wb(arg);
      END;
    END;
  END;
END;

{---------------------------------------------------------------------------}
{ Angegebenen Pass des Assemblers durchfhren                               }
{---------------------------------------------------------------------------}
PROCEDURE Assemble(Pass:INTEGER);
VAR F:TEXT;
    S,arg:STRING;
    i,j:INTEGER;
    p:PInsn;
    pm:PMap;
BEGIN
  IP := $100;
  Assign(F, ParamStr(1));
  Reset(F);
  WHILE NOT EOF(F) DO BEGIN
    Readln(F,S);
    Strip(S, s_Both);
    i:=1;
    WHILE (i<=Length(S)) AND (Pos(Upcase(S[i]),idchars)>0) DO Inc(i);
    IF (i<=Length(S)) AND (S[i]=':') THEN BEGIN
      IF Pass=1 THEN AddSymbol(Copy(S,1,i-1), IP) ELSE BEGIN
        pm:=Mapper(Upstr(Copy(S,1,i-1)));
        IF (pm=NIL) OR (pm^.Ofs<>IP) THEN BEGIN
          Writeln('Assembler error: Symbol ',Copy(S,1,i-1),' different in pass 1 and 2:');
          IF pm=NIL THEN Write('  non-existant') ELSE
           Write('  at ',pm^.Ofs);
          Writeln(' in pass 1, at ',IP,' in pass 2');
        END;
      END;
      Delete(S,1,i);
      Strip(s,s_Both);
    END;
    i:=0;
    REPEAT
      Inc(i);
    UNTIL (i>Length(S)) OR (S[i]=' ') OR (S[i]=#9);
    arg:=Copy(S,i,255);
    IF i<Length(S) THEN S[0]:=Chr(i);
    Strip(arg, s_Both);
    Strip(S, s_Both);
    S:=Upstr(S);
    IF (S='') OR (S[1]=';') THEN ELSE
    IF S='DB' THEN HandleDB(arg,pass) ELSE
    IF S='SPACE' THEN HandleSpace(arg) ELSE
    IF S='DW' THEN HandleDW(arg,pass) ELSE BEGIN
      p:=Identify(s,arg);
      IF p=NIL THEN Writeln('Can''t match insn ',s,' ',arg) ELSE
       AssembleInsn(p,pass);
    END;
  END;
  Close(F);
END;

{---------------------------------------------------------------------------}
{ Programmcode ausgeben                                                     }
{---------------------------------------------------------------------------}
PROCEDURE Output;
VAR F:FILE;
BEGIN
  Assign(F,ParamStr(2));
  Rewrite(F,1);
  Blockwrite(F,buffer,IP-$100);
  CLose(F);
END;

{---------------------------------------------------------------------------}
{ MAP-Datei ausgeben                                                        }
{---------------------------------------------------------------------------}
PROCEDURE Mapfile;
VAR F:TEXT;
    p:PMap;
BEGIN
  IF ParamStr(3)='' THEN Exit;
  Assign(F,ParamStr(3));
  Rewrite(F);
  p:=Map;
  WHILE p<>NIL DO BEGIN
    Writeln(F, p^.Ofs, ' ', p^.Name);
    p:=p^.Next;
  END;
  Close(F);
END;

BEGIN
  Writeln('Z80 Assembler');
  IF (ParamCount<2) OR (ParamCount>3) THEN BEGIN
    Writeln('Usage: Z80ASM source.asm target.com [target.map]');
    Halt;
  END;
  LoadInsns;
  Assemble(1);
  Assemble(2);
  Output;
  Mapfile;
END.
