{---------------------------------------------------------------------------}
{                                                                           }
{   Symbolischer Disassembler                                               }
{                                                                           }
{   Dieser Disassembler wird durch eine Tabelle in INSN.LST gesteuert.      }
{   Ein Befehl ist entweder ein ASCII-Text mit Platzhaltern, oder ein       }
{   Statusbergang (fr die Befehle mit Prfix)                             }
{                                                                           }
{   (c) copyright 1999 Stefan Reuther <Streu@gmx.de>                        }
{                                                                           }
{---------------------------------------------------------------------------}
UNIT Z80decode;

INTERFACE

{=== Symbolnamen ===}
TYPE PMap = ^TMap;
     TMap = RECORD
         Next:PMap;
         Name:STRING[63];
         Ofs:WORD;
     END;

CONST Map:PMap = NIL;

TYPE Str7 = STRING[7];
     Getter = FUNCTION(w:WORD):BYTE;

VAR Code:STRING;
    Adr:WORD;
    Disasm:STRING;

CONST SymbolOffset:WORD = 0;

PROCEDURE LoadInsns;
PROCEDURE LoadMap(VAR F:TEXT);
FUNCTION DecodeInsn(a:WORD; g:Getter):WORD;

IMPLEMENTATION

VAR Get:Getter;
    Comment:STRING;

{---------------------------------------------------------------------------}
{ Utilities                                                                 }
{---------------------------------------------------------------------------}
CONST Hex:ARRAY[0..15] OF CHAR='0123456789ABCDEF';

FUNCTION HexByte(y:BYTE):Str7;
BEGIN
  HexByte := Hex[y DIV 16] + Hex[y MOD 16];
END;

FUNCTION HexWord(y:WORD):Str7;
BEGIN
  HexWord := HexByte(y DIV 256) + HexByte(y MOD 256);
END;

{---------------------------------------------------------------------------}
{ Speicherzugriff                                                           }
{---------------------------------------------------------------------------}
FUNCTION GetByte:BYTE;
VAR y:BYTE;
    R:WORD;
BEGIN
  y := Get(Adr);
  Code:=Code + HexByte(y);
  GetByte:=y;
  Inc(Adr);
END;

FUNCTION GetWord:WORD;
VAR w:WORD;
BEGIN
  w:=GetByte;
  GetWord:=GetByte*256 + w;
END;

CONST RegNames:ARRAY[0..7] OF STRING[4] = ('B', 'C', 'D', 'E', 'H', 'L', '(HL)', 'A');
      ddRegs:ARRAY[0..3] OF STRING[2] = ('BC','DE','HL','SP');
      ppRegs:ARRAY[0..3] OF STRING[2] = ('BC','DE','??','SP');
      qqRegs:ARRAY[0..3] OF STRING[2] = ('BC','DE','HL','AF');

TYPE PInsn = ^TInsn;
     TInsn = RECORD
         Next:PInsn;
         Mask,Res:BYTE;
         Status:LONGINT;        { bit0 = initial, bit1 = 'a', ... }
         Action:STRING[31];
     END;

CONST Insns : PInsn = NIL;

{---------------------------------------------------------------------------}
{ Statusbits aus Zeichenfolge berechnen                                     }
{ @ = bit 0                                                                 }
{ A = bit 1    bis                                                          }
{ Z = bit 26                                                                }
{---------------------------------------------------------------------------}
FUNCTION NewStatus(S:STRING):LONGINT;
VAR L:LONGINT;
    c:CHAR;
BEGIN
  L:=0;
  REPEAT
    IF (S='') OR (S[1]=']') THEN BEGIN
      NewStatus:=L;
      Exit;
    END;
    c := Upcase(S[1]);
    Delete(S,1,1);
    IF (c>='@') AND (c<='Z') THEN L:=L OR (LONGINT(1) SHL (Ord(c)-Ord('@')));
  UNTIL FALSE;
END;

{---------------------------------------------------------------------------}
{ INSN.LST laden                                                            }
{---------------------------------------------------------------------------}
PROCEDURE LoadInsns;
VAR F:TEXT;
    S:STRING;
    Status:LONGINT;
    i,j,k:INTEGER;
    this:PInsn;
BEGIN
  S:=ParamStr(0);
  WHILE (S<>'') AND (S[Length(S)]<>':') AND (S[Length(S)]<>'\') DO Dec(S[0]);
  Assign(F, S+'insn.lst');
 {$I-}
  Reset(F);
 {$I+}
  IF IOresult<>0 THEN BEGIN
    Writeln('Could not load disassembler data: symbolic listing will be unavailable.');
    Exit;
  END;
  Status := 1;
  WHILE NOT EOF(F) DO BEGIN
    Readln(F,S);
    IF (S<>'') AND (S[1]<>';') THEN BEGIN
      IF S[1]='[' THEN BEGIN
        Status:=NewStatus(S);
      END ELSE BEGIN
        this := New(PInsn);
        this^.Status := Status;
        this^.Mask := 0;
        this^.Res := 0;
        FOR i:=1 TO 8 DO BEGIN
          IF S='' THEN BEGIN
            Writeln('Error in insn definition file');
            Halt;
          END;
          this^.Res := this^.Res*2;
          this^.Mask := this^.Mask*2;
          IF S[1]='1' THEN Inc(this^.Res);
          IF S[1]<>'.' THEN Inc(this^.Mask);
          Delete(S,1,1);
        END;
        WHILE (S<>'') AND (S[1]=' ') DO Delete(s,1,1);
        this^.Action:=S;
        this^.Next:=insns;
        insns:=this;
      END;
    END;
  END;
END;

{---------------------------------------------------------------------------}
{ Identifiziere Opcode in gegebenem Status                                  }
{---------------------------------------------------------------------------}
FUNCTION Identify(status:BYTE; Opcode:BYTE):PInsn;
VAR Mask:LONGINT;
    p:PInsn;
    found:PInsn;
BEGIN
  Mask := LONGINT(1) SHL Status;
  p:=insns;
  found:=NIL;
  WHILE p<>NIL DO BEGIN
    IF (p^.Status AND Mask<>0) AND (Opcode AND p^.Mask=p^.Res) THEN BEGIN
      IF (found=NIL) OR (p^.Mask > found^.Mask) THEN found:=p;
    END;
    p:=p^.Next;
  END;
  Identify:=found;
END;

{---------------------------------------------------------------------------}
{ Adresse auf Name abbilden                                                 }
{---------------------------------------------------------------------------}
FUNCTION Mapper(w:WORD):PMap;
VAR p:PMap;
BEGIN
  p:=Map;
  WHILE (p<>NIL) AND (p^.Ofs<>w) DO p:=P^.Next;
  Mapper:=p;
END;

{---------------------------------------------------------------------------}
{ Einen Befehl decodieren                                                   }
{---------------------------------------------------------------------------}
PROCEDURE DisassembleInsn;
VAR Opcode:BYTE;
    rStatus:BYTE;
    vStatus:BYTE;
    p:PInsn;
    i,j:INTEGER;
    S:STRING[7];
    w:WORD;
    pm:PMap;
BEGIN
  rStatus := 0;
  vStatus := 0;
  Code:='';
  REPEAT
    Opcode := GetByte;
    p := Identify(rStatus, Opcode);
    IF p=NIL THEN BEGIN
      Disasm:='?? <invalid>';
      Exit;
    END;
    i:=1;
   {--- Aktion ist Statusbergang ---}
    IF p^.Action[1]='>' THEN BEGIN
      IF p^.Action[2]='>' THEN rStatus:=Ord(Upcase(p^.Action[3]))-Ord('@')
      ELSE BEGIN
        rStatus:=Ord(Upcase(p^.Action[2]))-Ord('@');
        vStatus:=rStatus;
      END;
    END ELSE BEGIN
     {--- normaler Befehl ---}
      Disasm:='';
      i:=1;
      WHILE i<=Length(p^.Action) DO BEGIN
        IF p^.Action[i]='%' THEN BEGIN
          Inc(i);
          CASE p^.Action[i] OF
            '%':Disasm:=Disasm+'%';
            'r':Disasm:=Disasm+RegNames[(Opcode SHR 3) AND 7];
            'R':Disasm:=Disasm+RegNames[ Opcode        AND 7];
            'v':BEGIN
              j:=GetByte;
              IF j>=128 THEN Str(j-256,S) ELSE
              IF j=0 THEN S:='' ELSE BEGIN
                Str(j,S);
                S:='+'+S;
              END;
              Disasm:=Disasm+S;
            END;
            'j':BEGIN
              j:=GetByte;
              IF j>=128 THEN Disasm:=Disasm+HexWord(Adr + j - 256)+'h' ELSE
               Disasm:=Disasm+HexWord(Adr + j)+'h';
            END;
            'x':Disasm:=Disasm+Chr(Ord('@')+vStatus);
            'i':Disasm:=Disasm+HexByte(GetByte)+'h';
            'w':BEGIN
              w := GetWord;
              Disasm:=Disasm+HexWord(w)+'h';
              pm:=Mapper(w - SymbolOffset);
              IF pm<>NIL THEN BEGIN
                Comment := pm^.Name;
                IF SymbolOffset<>0 THEN Comment := Comment + ' (+'+HexWord(SymbolOffset) + 'h)';
              END;
            END;
            'd':Disasm:=Disasm+ddRegs[(Opcode SHR 4) AND 3];
            'p':Disasm:=Disasm+ppRegs[(Opcode SHR 4) AND 3];
            'q':Disasm:=Disasm+qqRegs[(Opcode SHR 4) AND 3];
            'b':Disasm:=Disasm+Hex[(Opcode SHR 3) AND 7];
          END;
        END ELSE BEGIN
          Disasm:=Disasm + p^.Action[i];
        END;
        Inc(i);
      END;
      Exit;
    END;
  UNTIL FALSE;
END;

{---------------------------------------------------------------------------}
{ MAP-Datei laden                                                           }
{---------------------------------------------------------------------------}
PROCEDURE LoadMap(VAR F:TEXT);
VAR a:WORD;
    S:STRING[63];
    P:PMap;
BEGIN
  WHILE NOT EOF(F) DO BEGIN
   {$I-}
    Readln(F,a,s);
   {$I+}
    IF IOResult<>0 THEN BEGIN
      Writeln('MAP file garbled - ignored rest of file.');
      Exit;
    END;
    WHILE (s<>'') AND (S[1]<=' ') DO Delete(s,1,1);
    New(P);
    P^.Next:=Map;
    Map:=P;
    P^.Ofs:=a;
    P^.Name:=S;
  END;
END;

{---------------------------------------------------------------------------}
{ Befehl an Adresse /a/ decodieren, Zugriff auf Speicher ber /g/           }
{---------------------------------------------------------------------------}
FUNCTION DecodeInsn(a:WORD; g:Getter):WORD;
BEGIN
  Adr:=a;
  Get:=g;
  Comment:='';
  DisassembleInsn;
  WHILE Length(Code)<16 DO Code:=Code+' ';
  IF Comment<>'' THEN BEGIN
    WHILE Length(Disasm)<24 DO Disasm:=Disasm+' ';
    Disasm:=Disasm + '; ' + Comment;
  END;
  DecodeInsn:=Adr;
END;

END.